"""
Catalogue maker (:mod:`~harmonia.mapper.catalogue_maker`)
===========================================================================

Make discrete catalogues from observed or simulated realisations.

.. autosummary::

    SourceCatalogue
    RandomCatalogue
    SphericalFKPCatalogue
    spherical_indicator

|

"""
import logging
import warnings

import numpy as np
from nbodykit.lab import CSVCatalog, FKPCatalog, UniformCatalog

from harmonia.utils import normalise_vector


def spherical_indicator(cartesian_position, bounding_radius):
    """Indicate whether an object lies within the a spherical domain
    centred at the origin.

    Parameters
    ----------
    cartesian_position : float, array_like
        Object position in Cartesian coordinates.
    bounding_radius : float
        Bounding radius of the spherical domain.

    Returns
    -------
    bool :class:`numpy.ndarray`
        `True` if the objection position lies within the spherical domain.

    """
    return np.linalg.norm(cartesian_position, axis=-1) <= bounding_radius


class RandomCatalogue(UniformCatalog):
    """Uniform random catalogue.

    Notes
    -----
    Origin of the catalogue is at a corner of the catalogue box.

    Parameters
    ----------
    mean_density : float
        Desired mean particle number density (in cubic :math:`h`/Mpc).
    boxsize : float, array_like
        Catalogue boxsize (in Mpc/:math:`h`) as a scalar or a triple of
        scalars.
    seed : int or None, optional
        Random seed of the catalogue (default is `None`).

    """

    def __init__(self, mean_density, boxsize, seed=None, comm=None):

        super().__init__(mean_density, boxsize, seed=seed, comm=comm)

        self.attrs['BoxSize'] = [boxsize,] * 3 \
            if np.isscalar(boxsize) else boxsize
        self.attrs['nbar'] = mean_density

        if self.comm is None or self.comm.rank == 0:
            self.logger.debug(
                "%s generated with origin at the box centre.", self
            )

    def __str__(self):

        str_info = "nbar={}, boxsize={}, seed={}".format(
            self.attrs['nbar'], self.attrs['BoxSize'], self.attrs['seed']
        )

        return f"{self.__class__.__name__}({str_info})"


class SourceCatalogue(CSVCatalog):
    """Catalogue from external sources.

    Notes
    -----
    This assumes the external catalogue has a Cartesian coordinate system
    with the origin at a corner of the Cartesian box.

    Parameters
    ----------
    source_file : str
        Catalogue source file.
    headings : list of str
        Column headings of the catalogue file.  This must contain three
        columns 'x', 'y' and 'z' for holding the Cartesian coordinates.
    mean_density: float or None, optional
        If not `None` (default), this overrides default catalogue mean
        number density (in cubic :math:`h`/Mpc) (e.g. to evade integral
        constraint).
    boxsize : float or None, optional
        If not `None` (default), this overrides default catalogue boxsize
        (in Mpc/:math:`h`) in case of unit upscaling.
    upscale : float, optional
        Scaling factor for converting any length unit to Mpc/:math:`h`
        (default is 1.), e.g. ``upscale=1.e-3`` for converting
        Kpc/:math:`h` to Mpc/:math:`h`.
    offset : bool, optional
        If `True` (default is `False`), add the velocity columns to
        position columns (for e.g. redshift-space distortions).
    offset_upscale : float, optional
        Scaling factor for converting the velocity offset length unit to
        Mpc/:math:`h` (default is 1.e-3), e.g. ``offset_upscale=1.e-3``
        for converting Kpc/:math:`h` to Mpc/:math:`h`.  The velocity offset
        should include the redshift conversion factor (the conformal
        Hubble parameter).
    **kwargs
        Parameters (other than `path` and `names`) to be passed to
        :class:`nbodykit.source.catalog.CSVCatalog`.

    Notes
    -----
    The use of ``offset=True`` assumes the source catalogue contains
    velocity data columns labelled by, for instance, 'vx', for each of
    the position data columns, for instance, 'x'.

    """

    def __init__(self, source_file, headings, mean_density=None, boxsize=None,
                 upscale=1., offset=False, offset_upscale=1.e-3, **kwargs):

        super().__init__(str(source_file), headings, **kwargs)

        # WARNING: Unresolved MPI bug with `nbodykit`.
        if self.comm is not None and self.comm.size > 1:
            warnings.warn(
                "Beware of enabling multi-processing for data catalogues: "
                "unresolved MPI issue with 'nbodykit'."
            )

        self.attrs['source'] = source_file
        self.attrs['offset'] = offset

        if boxsize is not None:
            self.attrs['BoxSize'] = [boxsize,] * 3 if np.isscalar(boxsize) \
                else boxsize
        else:
            self.attrs['BoxSize'] = \
                np.max(self['Position']) - np.min(self['Position'])

        if mean_density is not None:
            self.attrs['nbar'] = mean_density
        else:
            self.attrs['nbar'] = self.size / np.prod(self.attrs['BoxSize'])

        self['Position'] = \
            self['x'][:, None] * [upscale, 0, 0] \
            + self['y'][:, None] * [0, upscale, 0] \
            + self['z'][:, None] * [0, 0, upscale]

        # Add redshift-space distortions.
        if offset:
            self['Position'] = (
                self['vx'][:, None] * [offset_upscale, 0, 0] \
                + self['vy'][:, None] * [0, offset_upscale, 0] \
                + self['vz'][:, None] * [0, 0, offset_upscale]
            ) * normalise_vector(self['Position'])

        if self.comm is None or self.comm.rank == 0:
            self.logger.debug("%s generated.", self)

    def __str__(self):

        str_info = "nbar={}, boxsize={}, source={}".format(
            self.attrs['nbar'], self.attrs['BoxSize'], self.attrs['source']
        )

        return f"{self.__class__.__name__}({str_info})"


class SphericalFKPCatalogue:
    """FKP-style paired catalogues in a spherical domain.

    Parameters
    ----------
    data_catalogue : :class:`~.SourceCatalogue` *or None, optional*
        Data catalogue.  Cannot be `None` (default) unless `source_file`
        and `source_kwargs` are provided as an alternative to read the
        catalogue source file.
    source_file : str or None, optional
        Catalogue source file path.  This can only be `None` (default) if
        `data_catalogue` is already provided.
    source_kwargs : dict or None, optional
        Parameters to pass to :class:`~.SourceCatalogue`.  This can only
        be `None` (default) if `data_catalogue` is already provided.
    random_catalogue : :class:`nbodykit.base.catalog.CatalogSource` *or None, optional*
        Random catalogue (default is `None`).
    contrast : float or None, optional
        Mean density contrast compared to the data catalogue used to
        generate a random catalogue (default is `None`).  Ignored if
        `random_catalogue` is provided.
    mask : callable or None, optional
        Any veto mask function to be applied to both the data and random
        catalogues.  Must be a function of 3-d Cartesian
        coordinates only, assuming the origin is at the centre of
        the catalogues.
    selection : callable or None, optional
        Any selection function (normalised to unity) to be applied to both
        the data and random catalogues.  Must be a function of 3-d
        Cartesian coordinates only, assuming the origin is at the centre of
        the catalogues.
    weight : callable or None, optional
        Any weight function to be applied to both the data and random
        catalogues.  Must be a function of 3-d Cartesian coordinates only,
        assuming the origin is at the centre of the catalogues.
    random_seed : int or None, optional
        Random seed of the random catalogue (default is `None`).
    apply_selection_as_veto : bool, optional
        If `True` (default is `False`), `selection` returning boolean
        values is applied to the 'Selection' column of the catalogues
        instead of the 'NZ' column (see `nbodykit` documentation for
        this peculiarity.)

    Attributes
    ----------
    catalogue_pair : |FKPCatalog|
        FKP-style paired catalogue.
    data_catalogue : :class:`~.SourceCatalogue`
        Data catalogue.
    random_catalogue : :class:`nbodykit.base.catalog.CatalogSource`
        Random catalogue.
    attrs : dict
        Attributes including initialisation parameters.

    Notes
    -----
    If `selection` is provided, it must be normalised to the overall mean
    number density of the data catalogue.


    .. |FKPCatalog| replace::

        :class:`nbodykit.algorithms.convpower.catalog.FKPCatalog`

    """

    _msg = {
        'boxsizes': \
            "Different boxsizes for %s and %s catalogues: %s and %s.",
        'inscribe': (
            "Bounding sphere larger than the %s catalogue: "
            "diameter %.1f > boxsize %s."
        ),
        'centre': (
            "Centred %s catalogue with centred positions in 'Location' column."
        ),
        'sphericalise': "Restricted %s catalogue to spherical domain.",
        'mask': "Applied mask veto to %s catalogue.",
        'selection': "Applied selection function to %s catalogue.",
        'weight': "Applied weighting scheme to %s catalogue.",
    }

    def __init__(self, radius, data_catalogue=None, source_file=None,
                 source_kwargs=None, random_catalogue=None, contrast=None,
                 mask=None, selection=None, weight=None, random_seed=None,
                 apply_selection_as_veto=False):

        self.logger = logging.getLogger(self.__class__.__name__)

        if data_catalogue is None:
            self.data_catalogue = SourceCatalogue(source_file, **source_kwargs)
        else:
            self.data_catalogue = data_catalogue
            try:
                source_file = self.data_catalogue.attrs['source']
            except (AttributeError, KeyError):
                source_file = None

        if random_catalogue is not None:
            self.random_catalogue = random_catalogue
            contrast = \
                random_catalogue.attrs['nbar'] / data_catalogue.attrs['nbar']
        elif contrast is not None:
            self.random_catalogue = RandomCatalogue(
                contrast * self.data_catalogue.attrs['nbar'],
                self.data_catalogue.attrs['BoxSize'],
                seed=random_seed
            )
        else:
            self.random_catalogue = None

        self.attrs = {
            'bounding_radius': radius,
            'source': source_file,
            'contrast': contrast,
        }

        self.catalogue_pair = self._initialise(
            mask, selection, weight, apply_selection_as_veto
        )

    def __str__(self):

        str_info = (
            "radius={bounding_radius}, source={source}, contrast={contrast}"
        ).format(**self.attrs)

        return f"{self.__class__.__name__}({str_info})"

    def _initialise(self, mask, selection, weight, apply_selection_as_veto):

        radius = self.attrs['bounding_radius']

        data_boxsize = self.data_catalogue.attrs['BoxSize']

        if self.random_catalogue is None:
            rand_boxsize = None
        else:
            rand_boxsize = self.random_catalogue.attrs['BoxSize']
            # Compare data and random catalogue boxsizes.
            if not np.allclose(data_boxsize, rand_boxsize):
                warnings.warn(
                    self._msg['boxsizes'],
                    'data', 'random',
                    np.around(data_boxsize, decimals=0),
                    np.around(rand_boxsize, decimals=0)
                )

        # Restrict catalogued in the spherical domain.
        for name, catalogue, boxsize in zip(
                ['data', 'random'],
                [self.data_catalogue, self.random_catalogue],
                [data_boxsize, rand_boxsize]
            ):

            if catalogue is None:
                break

            # Check the bounding sphere is entirely within the catalogue.
            if np.any(np.less(boxsize, 2 * radius)):
                self.logger.debug(
                    self._msg['inscribe'], name, 2 * radius, boxsize
                )

            # Centre the coordinate origin if not already centred.
            is_origin_at_corner = np.any(
                np.isclose(
                    np.min(np.abs(catalogue['Position']), axis=0)
                    / np.max(np.abs(catalogue['Position']), axis=0),
                    0., atol=1.e-2
                )
            )
            if is_origin_at_corner:
                catalogue['Location'] = \
                    catalogue['Position'] - np.divide(boxsize, 2)
                self.logger.debug(self._msg['centre'], name)
            else:
                catalogue['Location'] = catalogue['Position']

            # Sphericalise and set base selection values ('NZ') for the
            # `nbodykit` `ConvolvedFFTPower` algorithm.
            catalogue['Selection'] &= \
                spherical_indicator(catalogue['Location'], radius)
            self.logger.debug(self._msg['sphericalise'], name)

            # Apply any mask, selection or weight.
            catalogue['NZ'] = \
                self.data_catalogue.attrs['nbar'] * np.ones(catalogue.csize)

            if callable(mask):
                catalogue['Selection'] &= mask(catalogue['Location'])
                self.logger.debug(self._msg['mask'], name)

            if callable(selection):
                if apply_selection_as_veto:
                    catalogue['Selection'] *= selection(catalogue['Location'])
                else:
                    catalogue['NZ'] *= selection(catalogue['Location'])
                self.logger.debug(self._msg['selection'], name)

            if callable(weight):
                catalogue['Weight'] *= weight(catalogue['Location'])
                self.logger.debug(self._msg['weight'], name)

        # Construct FKP catalogue.
        catalogue_pair = FKPCatalog(self.data_catalogue, self.random_catalogue)

        return catalogue_pair



import numpy as np
import pytest

from harmonia.algorithms.discretisation import DiscreteSpectrum
from harmonia.algorithms.arrays import CartesianArray, SphericalArray

DISC_PARAMS = dict(
    radius=50.,
    condition='dirichlet',
    highcut=0.25,
    lowcut=0.,
    maxdeg=None,
    mindeg=0,
)

FIXTURE_PARAMS = dict(
    orders=[0, 2],
    wavenumbers=[0.001, 0.04, 0.1]
)


@pytest.fixture(scope='module')
def spherical_array():

    discrete_spectrum = DiscreteSpectrum(**DISC_PARAMS)

    array = SphericalArray(discrete_spectrum)

    # Arbitraty coefficient data entries for testing only.
    array[:] = np.array([
        index[0] + index[-1] * 1j
        for index in array.array['index']
    ])

    return array


class TestSphericalArray:

    @pytest.mark.parametrize(
        "pivot,collapse,value",
        [
            ('natural', 'mean', [0. + 1.j, 1. + 1.j, 2. + 1.j]),
            ('spectral', None, [0. + 1.j, 1. + 1.j, 1. + 1.j,])
        ]
    )
    def test_vectorise(self, spherical_array, pivot, collapse, value):
        assert spherical_array.vectorise(pivot, collapse)[:3] \
            == pytest.approx(value), \
            "Vectorisation of SphericalArray test instance failed."

    @pytest.mark.parametrize(
        "key,value",
        [
            (-1, 7. + 1.j),
            ((0, 0, 1), 0. + 1.j),
            (slice(-2, -1), 7. + 1.j)
        ]
    )
    def test___getitem__(self, spherical_array, key, value):
        assert spherical_array[key] == pytest.approx(value), \
            "Container emulation failed for SphericalArray test instance."

    @pytest.mark.parametrize(
        "key,value",
        [
            (-1, 100. - 100.j),
            ((0, 0, 1), 0. - 1.j),
            (slice(-2, -1), 0.5 + 1.5j)
        ]
    )
    def test___setitem__(self, spherical_array, key, value):
        spherical_array[key] = value
        assert spherical_array[key] == pytest.approx(value), \
            "Container emulation failed for SphericalArray test instance."


@pytest.fixture(scope='module')
def cartesian_array():

    array = CartesianArray(**FIXTURE_PARAMS)

    # Arbitraty power data entries for testing only.
    array[:] = np.concatenate([
        array.attrs['wavenumbers'] ** order
        for order in array.attrs['orders']
    ])

    return array


class TestCartesianArray:

    @pytest.mark.parametrize(
        "pivot,value",
        [
            ('order', [1., 1., 1., 1.e-6, 1.6e-3, 1.e-2]),
            ('wavenumber', [1., 1.e-6, 1., 1.6e-3, 1., 1.e-2])
        ]
    )
    def test_vectorise(self, cartesian_array, pivot, value):
        assert cartesian_array.vectorise(pivot) == pytest.approx(value), \
            "Vectorisation of SphericalArray test instance failed."

    @pytest.mark.parametrize(
        "key,value",
        [
            (-1, 0.01),
            ((0, 0.04), 1.),
            ('power_2', [1.e-6, 1.6e-3, 1.e-2])
        ]
    )
    def test___getitem__(self, cartesian_array, key, value):
        assert cartesian_array[key] == pytest.approx(value), \
            "Container emulation failed for CartesianArray test instance."

    @pytest.mark.parametrize(
        "key,value",
        [
            (-1, 0.02),
            ((0, 0.04), 2.),
            ('power_2', [2.e-6, 2.6e-3, 2.e-2])
        ]
    )
    def test___setitem__(self, cartesian_array, key, value):
        cartesian_array[key] = value
        assert cartesian_array[key] == pytest.approx(value), \
            "Container emulation failed for CartesianArray test instance."



"""
***************************************************************************
Cosmological modelling (:mod:`~harmonia.cosmology`)
***************************************************************************

Provide general and fiducial cosmological models and compute derived model
quantities.

Cosmological models are implemented by `nbodykit` (see `documentation
<https://nbodykit.readthedocs.io>`_).

.. autosummary::

    BaseModel

|

"""
import logging
from pprint import pformat

from nbodykit.cosmology import Cosmology

from .geometry import differential_AP_distortion, redshift_from_distance
from .scale_dependence import (
    modified_power_spectrum,
    scale_dependence_modification,
    scale_dependent_bias,
)


class BaseModel(Cosmology):
    """Fixed (fiducial) cosmological model built with a parameter
    dictionary read from an external file, e.g. with file contents

    .. code-block::

        {
            'h': 0.70,
            'Omega0_b': 0.044,
            'Omega0_cdm': 0.226,
            'sigma8': 0.8,
        }

    Parameters
    ----------
    source_file : *str or* :class:`pathlib.Path`
        Cosmological parameter file (as a Python dictionary).
    comm : :class:`mpi4py.Comm` *or None*, optional
        MPI communicator (default is `None`).

    """

    def __init__(self, source_file, comm=None):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm

        with open(source_file) as cosmo_source:
            source_params = {
                k: round(v, ndigits=8)
                for k, v in eval(cosmo_source.read()).items()
            }

        cosmo_params = {
            param_name: param_value
            for param_name, param_value in source_params.items()
            if param_name != 'sigma8'
        }

        super().__init__(**cosmo_params)

        self.match(sigma8=source_params['sigma8'])

        if self.comm is None or self.comm.rank == 0:
            self.logger.info(
                "Created cosmological model with parameters:\n%s.",
                pformat(source_params)
            )



"""
Map transform (:mod:`~harmonia.mapper.map_transform`)
===========================================================================

Transform discrete catalogues to Fourier-space map data.

.. autosummary::

    SphericalMap
    CartesianMap

|

"""
import logging

import numpy as np
from nbodykit.lab import ConvolvedFFTPower

from harmonia.algorithms.discretisation import DiscreteSpectrum
from harmonia.algorithms.arrays import CartesianArray, SphericalArray
from harmonia.algorithms.bases import spherical_besselj, spherical_harmonic
from harmonia.surveyor.coordinates import cartesian_to_spherical as c2s
from harmonia.utils import Progress


class SphericalMap:
    r"""Discretised spherical Fourier map from catalogue sources.

    Parameters
    ----------
    catalogues : :class:`~.catalogue_maker.SphericalFKPCatalogue`
        FKP-style paired catalogues in a spherical domain.
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum for the map.
    initialise : bool, optional
        If `True` (default), map transform is performed upon creation.

    Attributes
    ----------
    catalogues : :class:`~.catalogue_maker.SphericalFKPCatalogue`
        FKP-style paired catalogues associated with the map.
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum associated with the map.
    attrs : dict
        Attributes inherited upon creation.

    """

    def __init__(self, catalogues, disc, initialise=True):

        self.logger = logging.getLogger(self.__class__.__name__)

        self.catalogues = catalogues
        self.disc = disc

        self.attrs = {}
        self.attrs.update(self.catalogues.attrs)
        self.attrs.update(self.disc.attrs)

        # Coefficient directories which internally store the
        # transformed results.
        self._data_coeff = {}
        self._rand_coeff = {}

        self._density_contrast = None
        self._mode_power = None

        if initialise:
            self._density_contrast = self._transform()

    def __str__(self):

        str_info = "{}, source={}, contrast={}".format(
            self.disc, self.attrs['source'], self.attrs['contrast']
        )

        return f"{self.__class__.__name__}({str_info})"

    def __getstate__(self):

        state = {
            'attrs': self.attrs,
            'disc': self.disc.__getstate__(),
            'data_coeff': self._data_coeff,
            'rand_coeff': self._rand_coeff,
            'density_contrast': self.density_contrast.__getstate__(),
            'mode_power': self.mode_power,
        }

        return state

    def __setstate__(self, state):

        for attr, val in state.items():
            if attr == 'disc':
                self.disc = DiscreteSpectrum._from_state(state['disc'])
            elif attr == 'density_contrast':
                self.density_contrast = \
                    SphericalArray._from_state(state['density_contrast'])
            elif attr.endswith('_coeff'):
                setattr(self, '_' + attr, val)
            else:
                setattr(self, attr, val)

    @classmethod
    def _from_state(cls, state):

        self = object.__new__(cls)

        self.__setstate__(state)

        # pylint: disable=protected-access
        if not hasattr(self, '_mode_power'):
            self._mode_power = None
        if not hasattr(self, '_data_coeff'):
            self._data_coeff = {}
        if not hasattr(self, '_rand_coeff'):
            self._rand_coeff = {}

        return self

    @property
    def density_contrast(self):
        """Spherical Fourier coefficients for the density contrast between
        data and random catalogues.

        Notes
        -----
        When the spherical map is initialsed upon creation or this
        is directly or indirectly accessed without initialising the
        spherical map upon creation, discrete spherical Fourier transform
        is performed by direct summation over all selected objects in the
        paired catalogues.

        Calling this method stores the transformed number densities of
        data and random catalogues internally for further processing,
        e.g. for :meth:`mode_power`.  Computational redundancy is reduced
        by employing parity relations between spherical harmonics of the
        same degree but opposite orders.

        Returns
        -------
        :class:`~harmonia.algorithms.arrays.SphericalArray`
            Density contrast coefficients of the spherical map.

        """
        if self._density_contrast is not None:
            return self._density_contrast

        self._density_contrast = self._transform()

        return self._density_contrast

    @density_contrast.setter
    def density_contrast(self, value):

        if isinstance(value, SphericalArray):
            self._density_contrast = value
        else:
            raise TypeError(
                "Only SphericalArray objects can be set as `density_contrast`."
            )

    @property
    def mode_power(self):
        """Spherical Fourier mode power suitably normalised.

        In the simplest case of a full-sky statistically isotropic map
        without any mode coupling, this is equivalent to the Cartesian
        power spectrum at the same wavenumbers.

        Returns
        -------
        dict
            Spherical mode power at discrete mode wavenumbers with mode
            counts and indices.

        """
        if self._mode_power is not None:
            return self._mode_power

        # Sort all variables by wavenumber.
        sort_order = np.argsort(
            list(self.disc.wavenumbers.values())
        )

        wavenumbers = np.array(
            list(self.disc.wavenumbers.values())
        )[sort_order]

        normalisations = np.array(
            list(self.disc.normalisations.values())
        )[sort_order]

        mode_indices = [
            list(self.disc.wavenumbers.keys())[order_index]
            for order_index in sort_order
        ]

        mode_counts = np.zeros_like(wavenumbers)
        mode_powers = np.zeros_like(wavenumbers)

        # Filter the spherical structure array by wavenumber and flatten.
        for idx, k in enumerate(wavenumbers):
            selector = (self.density_contrast.array['wavenumber'] == k)
            subarray = self.density_contrast.array['coefficient'][selector]
            mode_counts[idx] = np.sum(selector)
            mode_powers[idx] = \
                normalisations[idx] * np.mean(np.abs(subarray) ** 2)

        return {
            'wavenumbers': wavenumbers,
            'mode_indices': mode_indices,
            'mode_counts': mode_counts,
            'mode_powers': mode_powers,
        }

    def _transform(self):

        density_contrast = SphericalArray(self.disc)

        if not self._data_coeff or not self._rand_coeff:  # if either is empty
            progress = Progress(
                density_contrast.size,
                process_name="spherical transform",
                logger=self.logger
            )

            self.logger.info("Transforming %s...", self)

            # Transform degree by degree and update the coefficient directory.
            for deg_idx, deg in enumerate(self.disc.degrees):
                self._transform_degree(deg)
                progress.report(sum(self.disc.mode_counts[:(deg_idx + 1)]) - 1)

            self.logger.info("... %s transformed.", self)

        self.logger.info("Computing density contrast for %s...", self)

        # Subtract random catalogue coefficients and normalise by data
        # catalogue mean number density to obtain density contrast.
        for mode_pos, mode_index in enumerate(density_contrast.array['index']):
            density_contrast[mode_pos] = (
                self._data_coeff[tuple(mode_index)]
                - self._rand_coeff[tuple(mode_index)]
            ) / self.catalogues.data_catalogue.attrs['nbar']

        self.logger.info("... computed density contrast for %s.", self)

        return density_contrast

    def _transform_degree(self, ell):

        nmax = self.disc.depths[self.disc.degrees.index(ell)]

        loc_data = c2s(self.catalogues.data_catalogue['Location'])
        loc_rand = c2s(self.catalogues.random_catalogue['Location'])
        vet_data = self.catalogues.data_catalogue['Selection']
        vet_rand = self.catalogues.random_catalogue['Selection']
        sel_data = self.catalogues.data_catalogue['NZ'] \
            / self.catalogues.data_catalogue.attrs['nbar']
        sel_rand = self.catalogues.random_catalogue['NZ'] \
            / self.catalogues.data_catalogue.attrs['nbar']
        wgt_data = self.catalogues.data_catalogue['Weight']
        wgt_rand = self.catalogues.random_catalogue['Weight']

        coeff_data, coeff_rand = {}, {}

        # Unpack spherical coordinates.
        r_data, t_data, p_data = loc_data.swapaxes(0, 1)[:]
        r_rand, t_rand, p_rand = loc_rand.swapaxes(0, 1)[:]

        for m in range(- ell, 1):
            for n in range(1, nmax + 1):
                k = self.disc.wavenumbers[(ell, n)]
                coeff_data[(ell, m, n)] = complex(
                    np.sum(
                        vet_data[:] * sel_data[:] * wgt_data[:]
                        * spherical_besselj(ell, k * r_data)
                        * spherical_harmonic(ell, m, t_data, p_data, conj=True)
                    )
                )
                coeff_rand[(ell, m, n)] = complex(
                    np.sum(
                        vet_rand[:] * sel_rand[:] * wgt_rand[:]
                        * spherical_besselj(ell, k * r_rand)
                        * spherical_harmonic(ell, m, t_rand, p_rand, conj=True)
                    ) / self.catalogues.attrs['contrast']
                )

        # Employ parity relations for positive spherical orders.
        for m in range(1, ell + 1):
            for n in range(1, nmax + 1):
                coeff_data[(ell, m, n)] = (-1) ** m * np.conj(
                    coeff_data[(ell, - m, n)]
                )
                coeff_rand[(ell, m, n)] = (-1) ** m * np.conj(
                    coeff_rand[(ell, - m, n)]
                )

        self._data_coeff.update(coeff_data)
        self._rand_coeff.update(coeff_rand)


class CartesianMap:
    """Compressed Cartesian Fourier-space map from catalogue sources.

    Parameters
    ----------
    catalogues : :class:`~.SphericalFKPCatalogue` or :class:`FKPCatalog`
        FKP-style paired catalogues in a spherical domain.
    orders : sequence of int
        Orders of the power spectrum multipoles.
    kmin : float, optional
        Minimum wavenumber of the compressed map (in :math:`h`/Mpc)
        (default is `None`).
    kmax : float or None, optional
        Maximum wavenumber of the compressed map (in :math:`h`/Mpc)
        (default is `None`).
    dk : float or None, optional
        Wavenumber bin width (in :math:`h`/Mpc) (default is `None`).
    num_mesh : int, optional
        Mesh number per dimension for interpolating the discrete
        catalogues on a grid.
    resampler : {'cic', 'tsc', 'pcs'}, optional
        Grid assignment scheme (default is 'tsc') for catalogue
        interpolation.
    interlaced : bool, optional
        If `True` (default), use interlacing for aliasing mitigation.

    Attributes
    ----------
    catalogues : :class:`~.catalogue_maker.SphericalFKPCatalogue`
        FKP-style paired catalogues associated with the map.
    mesh : :class:`nbodykit.base.mesh.MeshSource`
        FFT mesh of the interpolated catalogues.
    power_multipoles : :class:`~harmonia.algorithms.arrays.CartesianArray`
        Power spectrum multipoles of different Legendre orders for
        wavenumber bins.

    """

    def __init__(self, catalogues, orders, kmin=None, kmax=None, dk=None,
                 num_mesh=256, resampler='tsc', interlaced=True):

        self.logger = logging.getLogger(self.__class__.__name__)

        self.catalogues = catalogues

        self.attrs = {arg: val for arg, val in locals().items() if arg != 'dk'}
        self.attrs.update(catalogues.attrs)

        mesh_kwargs = dict(
            Nmesh=num_mesh, resampler=resampler, interlaced=interlaced,
            compensated=True
        )
        try:  # deal with `SphericalFKPCatalogue`
            self.mesh = catalogues.catalogue_pair.to_mesh(**mesh_kwargs)
        except AttributeError:  # deal with `nbodykit`'s `FKPCatalog`
            self.mesh = catalogues.to_mesh(**mesh_kwargs)

        self.logger.debug("Mesh painted on a grid for %s.", catalogues)

        kmin = kmin or 1.e-4

        self.power_multipoles = self._compress(orders, kmin, kmax, dk)

    def __str__(self):

        str_info = "source={}, orders={}, scales=[{},{}], contrast={}".format(
            self.attrs['source'], self.attrs['orders'],
            self.attrs['kmin'], self.attrs['kmax'],
            self.attrs['contrast']
        )

        return f"{self.__class__.__name__}({str_info})"

    def __getstate__(self):

        state = {
            'attrs': self.attrs,
            'power_multipoles': self.power_multipoles,
        }

        return state

    def __setstate__(self, state):

        for attr, val in state.items():
            setattr(self, attr, val)

    @classmethod
    def _from_state(cls, state):  # internal classmethod

        self = object.__new__(cls)

        self.__setstate__(state)

        return self

    def _compress(self, orders, kmin, kmax, dk):

        results = ConvolvedFFTPower(
            self.mesh, poles=orders, kmin=kmin, kmax=kmax, dk=dk
        ).poles

        # Remove spurious leading bins from `nbodykit`.
        valid_bins = ~np.equal(results['modes'], 0) \
            & ~np.equal(results['modes'], 1)

        wavenumbers = results['k'][valid_bins]
        mode_counts = results['modes'][valid_bins]
        shot_noise = results.attrs['shotnoise']

        multipoles = CartesianArray(
            orders, wavenumbers,
            mode_counts=mode_counts, shot_noise=shot_noise
        )

        self.logger.info(
            "Compressed catalogues into power spectrum multipoles."
        )

        # Fill in Cartesian structure array.
        multipoles[:] = np.concatenate(
            [results[f'power_{order}'][valid_bins].real for order in orders]
        )

        return multipoles



r"""
Spherical couplings (:mod:`~harmonia.reader.couplings`)
===========================================================================

Compute spherical Fourier couplings coefficients for given survey
spefications and cosmological models.

.. autosummary::

    SphericalCoefficientWarning
    Couplings


Kernels
---------------------------------------------------------------------------

Coupling kernels are integrands without the coordinate Jacobian, which may
include the following factors:

    * angular mask :math:`M(\hat{\mathbf{r}})`;
    * radial selection :math:`\phi(r)`;
    * radial weight :math:`w(r)` and its derivative :math:`w'(r)`;
    * clustering evolution, :math:`D(z)`, which is the linear growth
      factor normalised to unity at the :math:`z = 0` epoch;
    * linear bias + clustering evolution
      :math:`G(z, k) = b(z, k) D(z) / b(z_*, k) D(z_*)` normalised to unity
      at a fiducial epoch :math:`z_*`, where :math:`b(z, k)` is the
      scale-dependent linear bias;
    * linear growth rate + clustering evolution
      :math:`F(z) = f(z) D(z) / f(z_*) D(z_*)` normalised to unity at a
      fiducial epoch :math:`z_*`, where :math:`f(z)` is the linear
      growth rate;
    * differential Alcock--Paczynski (AP) distortion

      .. math::

          \gamma(z) = \frac{
              \operatorname{d}\!\breve{\chi}(z)
          }{
              \operatorname{d}\!\chi(z)
          } \,,

      where :math:`\breve{r} = \breve{\chi}(z)` is the distance converted
      from redshift in a fiducial cosmology rather than from the true
      comoving distance--redshift correspondence :math:`z = \chi^{-1}(r)`.


Couplings
---------------------------------------------------------------------------

Coupling coefficients are computed by integrating the angular, radial and
RSD coupling kernels

.. math::

   \begin{align*}
       M_{\mu\nu} &= \int \operatorname{d}^2\!\hat{\mathbf{r}} \,
           Y_{\ell_\mu m_\mu}^*(\hat{\mathbf{r}})
           M(\hat{\mathbf{r}})
           Y_{\ell_\nu m_\nu}(\hat{\mathbf{r}}) \,, \\
       \Phi_{\mu\nu} &= \kappa_{\ell_\nu n_\nu}
           \int \operatorname{d}\!r \, r^2 w(\breve{r})
           j_{\ell_\mu}(k_{\ell_\mu n_\mu} \breve{r})
           j_{\ell_\nu}(k_{\ell_\nu n_\nu} r)
           G(z(r), k_{\ell_\nu n_\nu}) \phi(r) \,, \\
       \Upsilon_{\mu\nu} &=
           \frac{\kappa_{\ell_\nu n_\nu}}{k_{\ell_\nu n_\nu}}
           \int \operatorname{d}\!r \, r^2
           \frac{\operatorname{d}\!}{\operatorname{d}\!\breve{r}}
           \left[ w(\breve{r}) j_{\ell_\mu}(k_{\ell_\mu n_\mu} \breve{r})
               \right]
           j'_{\ell_\nu}(k_{\ell_\nu n_\nu} r)
           \gamma(z(r)) F(z(r)) \phi(r) \,,
   \end{align*}

over the spherical volume element, where :math:`k_{\ell n}` is the
discrete wavenumber.

When there is no angular masking (i.e. :math:`M(\hat{\mathbf{r}})` is
constant), the coupling coefficients reduce to :math:`M_{\mu\nu} =
\delta_{\mu\nu}`; if in addition radial selection, weighting and
evolutionary effects are all absent and there is no AP correction, then
:math:`M_{\mu\nu} \Phi_{\mu\nu} = \delta_{\mu\nu}`.

|

"""
import logging
import warnings

import numpy as np

from harmonia.algorithms.discretisation import DiscreteSpectrum
from harmonia.algorithms.integration import (
    angular_integral,
    pixelated_angular_integral,
    radial_integral,
)
from harmonia.utils import mpi_compute, restore_warnings
from ._kernels import angular_kernel, radial_kernel, RSD_kernel


class SphericalCoefficientWarning(UserWarning):
    """Warning issued for poorly determined spherical coefficient.

    """


class Couplings:
    """Angular, radial and RSD coupling coefficients computed for given
    survey and cosmological specifications.

    Notes
    -----
    Survey specification functions must be given in spherical coordinates
    and may include the following:

        * 'mask' for angular mask;
        * 'selection', 'weight' and 'weight_derivative' for radial
          selection, weighting and weight derivative.

    Cosmological specification functions must be given in redshift
    and/or the radial coordinate and may include the following:

        * 'z_from_r' and 'chi_of_z' for cosmological comoving
          distance-to-redshift conversion and fiducial
          redshift-to-comoving distance conversion;
        * 'bias_evolution', 'growth_evolution' for linear bias and
          linear growth rate, normalised to unity at a fiducial epoch;
        * 'clustering_evolution' for clustering evolution as a
          function of the redshift normalised to unity at the
          current epoch;
        * 'differential_AP_distortion' for differential AP distortion
          as a function of the redshift.

    Parameters
    ----------
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum associated with the couplings.
    survey_specs : dict{str: callable or None} or None, optional
        Survey specification functions as detailed above.
    cosmo_specs : dict{str: callable or None} or None, optional
        Cosmological specification functions as detailed above.
    initialise : bool, optional
        If `True`, compile all coupling coefficients upon creation.
    external_angular_couplings : dict{tuple(tuple, tuple): complex} or None, optional
        Pre-compute angular couplings (default is `None`).
    pixelate : int or None, optional
        If not `None` (default), this sets the 'NSIDE' parameter of
        `healpy` pixelation for angular coupling integration.
    comm : :class:`mpi4py.MPI.Comm` *or None, optional*
        MPI communicator (default is `None`).

    Attributes
    ----------
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum associated with the couplings.
    couplings : dict{str: dict}
        Directory for all coupling coefficients of different
        coupling types.

    """

    _coupling_types = ['angular', 'radial', 'rsd']
    _survey_specs = dict.fromkeys([
        'mask', 'selection', 'weight', 'weight_derivative'
    ])

    _cosmo_specs = dict.fromkeys([
        'z_from_r', 'chi_of_z',
        'bias_evolution', 'growth_evolution',
        'clustering_evolution',
        'differential_AP_distortion',
    ])

    def __init__(self, disc, survey_specs=None, cosmo_specs=None,
                 initialise=True, external_angular_couplings=None,
                 pixelate=None, comm=None):

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm
        self.initialised = False

        self.disc = disc

        if survey_specs is not None:
            self._survey_specs.update(survey_specs)
        if cosmo_specs is not None:
            self._cosmo_specs.update(cosmo_specs)

        self.couplings = {
            coupling_type: {}
            for coupling_type in self._coupling_types
        }
        if external_angular_couplings is not None:
            self.load_angular_couplings(external_angular_couplings)
        if initialise:
            self.compile_couplings(pixelate=pixelate)

    def __getstate__(self):

        state = {
            'disc': self.disc.__getstate__(),
            'couplings': self.couplings,
        }

        return state

    def __setstate__(self, state):

        for attr, value in state.items():
            if attr == 'disc':
                self.disc = DiscreteSpectrum._from_state(state['disc'])
            else:
                # NOTE: Strip '_' for backward compatibility; may be
                # removed in the future.
                setattr(self, attr.strip('_'), value)

    def __getitem__(self, key):
        r"""Access coupling coefficient by key.

        Notes
        -----
        Only accessible if initialised with :meth:`compile_couplings`.
        The access key is a tuple specifying the coupling type and the
        pair of triplet/double indices, e.g.
        ``['angular', (0, 0), (3, -1, 1)]`` for :math:`M_{0,0,3,-1}`.

        Parameters
        ----------
        key : tuple
            Coefficient access key.  Must contain elements of types
            specified above.

        Returns
        -------
        float or complex
            Coupling coefficient.

        Raises
        ------
        AttributeError
            If the coupling coefficients have not been initialised.

        """
        if not self.initialised:
            raise AttributeError(
                "Unitialised state: coupling coefficients not compiled. "
                "Please call `compile_couplings` first."
            )

        coupling_type, mu, nu = key

        if coupling_type == 'angular':
            mu, nu = (mu[0], mu[1]), (nu[0], nu[1])
            # Use parity to return an angular coupling coefficient
            # that is not internally stored.
            if mu > nu:
                return np.conj(self.couplings['angular'][nu, mu])

            return self.couplings['angular'][mu, nu]

        mu, nu = (mu[0], mu[-1]), (nu[0], nu[-1])

        return self.couplings[coupling_type][mu, nu]

    def load_angular_couplings(self, angular_couplings):
        """Load pre-computed angular coupling coefficients which are
        independent of the cosmological model.

        Parameters
        ----------
        angular_couplings : dict{tuple(tuple, tuple): complex}
            Pre-compute angular couplings.

        Raises
        ------
        ValueError
            If the number of entries in `angular_couplings` do not match
            the class instance.

        """
        num_unordered_index_pair = sum(range(
            1, sum([2 * deg + 1 for deg in self.disc.degrees])
        ))
        num_ordered_index_pair = sum(
            [2 * deg + 1 for deg in self.disc.degrees]
        ) ** 2

        if len(angular_couplings) \
                not in [num_unordered_index_pair, num_ordered_index_pair]:
            raise ValueError(
                "Number of index pairs in loaded angular couplings "
                "do not match the `disc` attribute."
            )

        self.couplings.update({'angular': angular_couplings})

    def compile_couplings(self, pixelate=None):
        """Compile all coupling coefficients.

        Parameters
        ----------
        pixelate : int or None, optional
            If not `None` (default), this sets the 'NSIDE' parameter of
            `healpy` pixelation for angular coupling integration.

        """
        for coupling_type in self._coupling_types:
            # Only compile empty coupling directories.
            if not self.couplings[coupling_type]:
                self._compile_couplings_by_type(
                    coupling_type, pixelate=pixelate
                )

        self.initialised = True

    def compute_coefficient(self, mu, nu, coupling_type, pixelate=None):
        r"""Compute coupling coefficients for given triplet indices.

        Parameters
        ----------
        mu, nu : tuple(int, int) or tuple(int, int, int)
            Coefficient triplet or reduced doublet index.
        coupling_type : {'angular', 'radial', 'RSD'}
            Coupling type.
        pixelate : int or None, optional
            If not `None` (default), this sets the 'NSIDE' parameter of
            `healpy` pixelation for computing angular coupling
            coefficients.

        Returns
        -------
        cofficient : float or complex :class:`numpy.ndarray`
            Coupling coefficient of the specified type.

        """
        if coupling_type == 'angular':
            if not callable(self._survey_specs['mask']):  # Kronecker delta
                coefficient = complex(mu[0] == nu[0] and mu[1] == nu[1])
            else:
                if pixelate:
                    coefficient = pixelated_angular_integral(
                        lambda theta, phi: angular_kernel(
                            theta, phi, mu, nu,
                            mask=self._survey_specs['mask']
                        ), nside=pixelate
                    )
                else:
                    with warnings.catch_warnings(record=True) as any_warning:
                        coefficient = angular_integral(
                            lambda theta, phi: angular_kernel(
                                theta, phi, mu, nu,
                                mask=self._survey_specs['mask']
                            )
                        )
                    if any_warning and not np.isclose(coefficient, 0.):
                        warnings.warn(
                            "Poorly determined angular coupling coefficients "
                            "for index pair {} and {}.".format(mu, nu),
                            category=SphericalCoefficientWarning
                        )
            return coefficient

        k_mu = self.disc.wavenumbers[mu[0], mu[-1]]
        k_nu = self.disc.wavenumbers[nu[0], nu[-1]]
        kappa_nu = self.disc.normalisations[nu[0], nu[-1]]

        spec_funcs = {}
        checklists = {
            '_survey_specs': ['selection', 'weight'],
            '_cosmo_specs': ['z_from_r', 'chi_of_z', 'clustering_evolution'],
        }

        if coupling_type == 'radial':
            checklists['_cosmo_specs'].extend(['bias_evolution'])
            spec_funcs.update({
                spec: getattr(self, spec_type)[spec]
                for spec_type, spec_list in checklists.items()
                for spec in spec_list
            })

            if (mu[0] == nu[0]) and not any(map(callable, spec_funcs.items())):
                coefficient = float(mu[-1] == nu[-1])  # Kronecker delta
            else:
                with warnings.catch_warnings(record=True) as any_warning:
                    coefficient = kappa_nu * radial_integral(
                        lambda r: radial_kernel(
                            r, mu, nu, k_mu, k_nu, **spec_funcs
                        ),
                        self.disc.attrs['boundary_radius']
                    )
                if any_warning:
                    warnings.warn(
                        "Poorly determined radial coupling coefficients "
                        "for index pair {} and {}.".format(mu, nu),
                        category=SphericalCoefficientWarning
                    )
            return coefficient

        if coupling_type == 'rsd':
            checklists['_survey_specs'].extend(['weight_derivative'])
            checklists['_cosmo_specs'].extend(
                ['growth_evolution', 'differential_AP_distortion']
            )
            spec_funcs.update({
                spec: getattr(self, spec_type)[spec]
                for spec_type, spec_list in checklists.items()
                for spec in spec_list
            })

            with warnings.catch_warnings(record=True) as any_warning:
                coefficient = kappa_nu / k_nu * radial_integral(
                    lambda r: RSD_kernel(r, mu, nu, k_mu, k_nu, **spec_funcs),
                    self.disc.attrs['boundary_radius']
                )
            if any_warning:
                warnings.warn(
                    "Poorly determined RSD coupling coefficients "
                    "for index pair {} and {}.".format(mu, nu),
                    category=SphericalCoefficientWarning
                )

            return coefficient

        raise ValueError(f"Unknown coupling type: {coupling_type}.")

    def save(self, output_file):
        """Save compiled couplings as a .npz file.

        Parameters
        ----------
        output_file : *str or* :class:`pathlib.Path`
            Output file path.

        """
        np.savez(output_file, **self.__getstate__())

    @classmethod
    def load(cls, input_file, comm=None):
        """Load compiled couplings from a .npz file.

        Parameters
        ----------
        input_file : *str or* :class:`pathlib.Path`
            Input file path.

        """
        state_data = np.load(input_file, allow_pickle=True)
        state = {}
        for attr in state_data.files:
            try:
                state.update({attr: state_data[attr].item()})
            except ValueError:
                state.update({attr: state_data[attr]})

        self = object.__new__(cls)

        self.__setstate__(state)

        self.logger = logging.getLogger(self.__class__.__name__)
        self.comm = comm

        self.initialised = True

        return self

    def _compile_couplings_by_type(self, coupling_type, pixelate=None):

        indices_to_compile = self._gen_operable_indices(coupling_type)
        index_compiler = lambda index: self._compile_couplings_by_index(
            index, coupling_type, pixelate=pixelate
        )

        if self.comm is None or self.comm.rank == 0:
            self.logger.info(
                "Compiling %s couplings...", self._alias(coupling_type)
            )
        if self.comm is not None:
            self.comm.Barrier()

        with warnings.catch_warnings(record=True) as any_warnings:
            list_of_compiled_couplings_ = mpi_compute(
                indices_to_compile, index_compiler, comm=self.comm,
                process_name="{} couplings compilation".format(
                    self._alias(coupling_type)
                )
            )

        if any_warnings:
            restore_warnings(any_warnings)
            warnings.warn(
                "Poor coefficient determinations for {} couplings."
                .format(coupling_type),
                SphericalCoefficientWarning
            )

        for compiled_couplings in list_of_compiled_couplings_:
            self.couplings[coupling_type].update(compiled_couplings)

        if self.comm is not None:
            self.comm.Barrier()
        if self.comm is None or self.comm.rank == 0:
            self.logger.info(
                "... compiled %s couplings.", self._alias(coupling_type)
            )

    def _compile_couplings_by_index(self, fixed_index, coupling_type,
                                    pixelate=None):

        mu = (fixed_index[0], fixed_index[1]) if coupling_type == 'angular' \
            else (fixed_index[0], fixed_index[-1])

        # Reduce the number of indices to compile by exploting
        # exchange symmetry between indices.
        above_from = mu if coupling_type == 'angular' else None
        variable_indices = self._gen_operable_indices(
            coupling_type, above_from=above_from
        )

        couplings_for_index = {}
        for nu in variable_indices:
            couplings_for_index.update({
                (mu, nu): self.compute_coefficient(
                    mu, nu, coupling_type, pixelate=pixelate
                )
            })

        return couplings_for_index

    def _gen_operable_indices(self, coupling_type, above_from=None):

        if coupling_type == 'angular':
            operable_indices = [
                (ell, m)
                for ell in self.disc.degrees
                for m in range(- ell, ell + 1)
            ]
            if above_from is not None:
                operable_indices = [
                    index for index in operable_indices if index >= above_from
                ]
        else:
            operable_indices = [
                (ell, n)
                for ell, nmax in zip(self.disc.degrees, self.disc.depths)
                for n in range(1, nmax + 1)
            ]

        return operable_indices

    @staticmethod
    def _alias(coupling_type):

        return 'RSD' if coupling_type == 'rsd' else coupling_type


def _group_couplings(couplings):
    """Change a directory of couplings from nested dictionaries to a
    dictionary of arrays for mass access to raidial and RSD coupling
    coefficients of adjacent indices.

    Parameters
    ----------
    couplings : :class:`~.reader.couplings.Couplings`
        Couplings object (default is `None`).

    Returns
    -------
    grouped_couplings : dict{tuple: list}
        Radial and RSD couplings grouped for mass access.

    """
    disc = couplings.disc

    # pylint: disable=protected-access
    key_indices = couplings._gen_operable_indices('radial')

    grouped_couplings = {}
    for coupling_type in ['radial', 'rsd']:
        coupling_arrays_dir = {
            key_index: [
                np.asarray([
                    couplings[coupling_type, key_index, (ell, n)]
                    for n in range(1, nmax + 1)
                ])
                for ell, nmax in zip(disc.degrees, disc.depths)
            ]
            for key_index in key_indices
        }
        grouped_couplings.update({coupling_type: coupling_arrays_dir})

    return grouped_couplings



import numpy as np
import pytest

from harmonia.algorithms.bases import spherical_besselj as sph_jn
from harmonia.algorithms.integration import (
    angular_integral,
    pixelated_angular_integral,
    radial_integral,
)

from . import NamedFunction, display_mathematica_query as show_query


@pytest.fixture(scope='module')
def radial_func():
    return NamedFunction(
        "SphericalBesselJ[5, r] * D[SphericalBesselJ[10, r], r]",
        lambda r: sph_jn(5, r) * sph_jn(10, r, derivative=True)
    )


@pytest.fixture(scope='module')
def angular_func():
    return NamedFunction(
        "Sin[theta + phi]",
        lambda theta, phi: np.sin(theta + phi)
    )


@pytest.mark.parametrize(
    "rmax,value",
    [
        (10, 0.211562),
    ]
)
def test_radial_integral(radial_func, rmax, value):

    show_query(f"Integrate[r^2 {str(radial_func)}, {{r, 0, {rmax}}}]")
    assert radial_integral(radial_func, rmax) == pytest.approx(value), \
        "Incorrect radial integration results for the test function."


def test_angular_integral(angular_func):

    show_query(
        f"Integrate[{str(angular_func)}, {{theta, 0, Pi}}, {{phi, 0, 2 Pi}}]"
    )
    assert angular_integral(angular_func) == pytest.approx(0.), \
        "Incorrect angular integration results for the test function."


@pytest.mark.parametrize(
    "nside",
    [64, 128]
)
def test_pixelated_angular_integral(angular_func, nside):

    show_query(
        f"Integrate[{str(angular_func)}, {{theta, 0, Pi}}, {{phi, 0, 2 Pi}}]"
    )
    assert pixelated_angular_integral(angular_func, nside) \
        == pytest.approx(0.), \
        "Incorrect angular integration results for the test function."



"""
***************************************************************************
Map reader (:mod:`~harmonia.reader`)
***************************************************************************

Model Fourier-space map statistics and construct cosmological likelihoods.

.. note::

    Unless otherwise specified, the length dimension in the module is
    in units of Mpc/:math:`h`.

"""
from .couplings import Couplings
from .likelihoods import (
    LogLikelihood,
    cartesian_moments,
    complex_normal_pdf,
    modified_student_pdf,
    multivariate_normal_pdf,
    spherical_covariance,
)
from .models import CartesianMultipoles, SphericalCorrelator



"""
Structured arrays (:mod:`~harmonia.algorithms.arrays`)
===========================================================================

Provide structured arrays for cosmological data.

.. autosummary::

    DataArray
    SphericalArray
    CartesianArray

|

"""
from collections.abc import Sequence
from itertools import product

import numpy as np

from .discretisation import DiscreteSpectrum

try:
    import cPickle as pickle
except ModuleNotFoundError:
    import pickle


class IndexingError(IndexError):
    """Exception raised for unsupported slicing or indexing in
    `__getitem__` methods.

    """


class DataArray:
    """Abstract data array with save and load methods.

    """

    def __getstate__(self):

        raise NotImplementedError

    def __setstate__(self, state):

        raise NotImplementedError

    def save(self, output_file, file_extension):
        """Save the structured array.

        Parameters
        ----------
        output_file : *str or* :class:`pathlib.Path`
            Output file path.
        extension : {'.pkl', '.npz'}
            Output file extension.

        """
        if file_extension == '.pkl':
            with open(output_file, 'wb') as output_data:
                pickle.dump(self, output_data, protocol=-1)
        elif file_extension == '.npz':
            np.savez(output_file, **self.__getstate__())
        else:
            raise IOError(
                "Unwritable output file. "
                "The file extension must be either .npz or .pkl."
            )

    @classmethod
    def load(cls, input_file):
        """Load the structured array from a .npz or .pkl file.

        Parameters
        ----------
        input_file : *str or* :class:`pathlib.Path`
            Input file path.

        """
        try:
            extension = input_file.suffix
        except AttributeError:
            extension = input_file.rpartition(".")[-1]

        if extension.endswith('npz'):
            state_data = np.load(input_file, allow_pickle=True)

            state = {}
            for attr in state_data.files:
                try:
                    state.update({attr: state_data[attr].item()})
                except ValueError:
                    state.update({attr: state_data[attr]})

            self = object.__new__(cls)
            self.__setstate__(state)
        elif extension.endswith('pkl'):
            with open(input_file, 'rb') as input_data:
                self = pickle.load(input_data)
        else:
            raise IOError(
                "Unreadable input file. "
                "The file extension must be either .npz or .pkl."
            )

        return self


class SphericalArray(DataArray):
    r"""Structured array for spherically decomposed cosmological data.

    Array is initialised with a discrete spectrum of Fourier modes
    and consists of three fields: the 'index' field of
    :math:`(\ell, m_\ell, n_\ell)` triplets, the 'wavenumber' field
    of discrete :math:`k_{\ell n}`, and the 'coefficient' field of
    spherically decomposed data.

    Parameters
    ----------
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum associated with the structured array.

    Attributes
    ----------
    array : :class:`numpy.ndarray`
        Structured NumPy array.
    size : int
        Total number of elements in the array.  This should equal the sum
        of ``disc.mode_counts``.
    attrs : dict
        Attributes of the structured array inherited from `disc`.

    See Also
    --------
    :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`

    """

    # Class-wide ``numpy.dtype`` for the structured array and any
    # collapsed array.
    _dtype = np.dtype({
        'names': ['index', 'wavenumber', 'coefficient', '_position'],
        'formats': ['(3,)i4', 'f8', 'c16', 'i8'],
    })

    _dtype_collapsed = np.dtype({
        'names': ['index', 'wavenumber', 'coefficient', '_position'],
        'formats': ['(2,)i4', 'f8', 'c16', 'i8'],
    })

    def __init__(self, disc):

        self.disc = disc
        self.size = sum(disc.mode_counts)

        self.attrs = {
            attr: getattr(disc, attr)
            for attr in ['degrees', 'depths', 'wavenumbers', 'mode_counts']
        }

        self.array, self._directory = self._initialise_array()

    def __str__(self):

        return f"{self.__class__.__name__}({str(self.disc)})"

    def __getitem__(self, key):
        """Get the 'coefficient' field value(s).

        The access key can be an integer, a slice expression, a tuple of
        index triplet or a string, e.g. ``[-1]``, ``[:]``, ``[(0, 0, 1)]``
        or ``'degree_0'``.

        Parameters
        ----------
        key : int, slice, tuple(int, int, int) or str
            'coefficient' field access key.

        Returns
        -------
        complex
            'coefficient' field data entry.

        """
        position = self._find_position(key)

        return self.array['coefficient'][position]

    def __setitem__(self, key, value):
        """Set the 'coefficient' field value(s).

        Parameters
        ----------
        key : int, tuple of int or slice
            'coefficient' field access key.
        value : complex
            'coefficient' field data entry.

        See Also
        --------
        :meth:`.SphericalArray.__getitem__`

        """
        position = self._find_position(key)

        self.array['coefficient'][position] = value

    def __getstate__(self):

        state = self.__dict__

        state.update({'disc': self.disc.__getstate__()})

        return state

    def __setstate__(self, state):

        for attr, value in state.items():
            if attr == 'disc':
                self.disc = DiscreteSpectrum._from_state(value)
            else:
                setattr(self, attr, value)

        # NOTE: For backward compatibility; may be removed in the future.
        if '_position' not in self.array.dtype.names:
            self.array = np.lib.recfunctions.append_fields(
                self.array, '_position', list(range(self.size))
            )

    @classmethod
    def _from_state(cls, state):  # internal classmethod

        self = object.__new__(cls)
        self.__setstate__(state)

        return self

    def vectorise(self, pivot, collapse=None):
        r"""Returrn a data vector from the 'coefficient' field.

        Vectorisation is performed by *pivoting* in either of the following
        orders of precedence---

            * 'natural': ordered by :math:`(\ell, m, n)`;
            * 'spectral': ordered by :math:`(k_{\ell n}, m)`.

        Subarrays of equivalent :math:`(\ell, n)` may be further collapsed
        over spherical order :math:`m` by simple averaging or averaging in
        quadrature.

        Parameters
        ----------
        pivot : {'natural', 'spectral'}
            Pivot order for vectorisation.
        collapse : {None, 'mean', 'qaudrature'}, optional
            If not `None` (default), subarrays are collapsed over
            equivalent spherical order :math:`m` by averaging ('mean') or
            averaging in quadrature ('qaudrature').

        Returns
        -------
        vectorised_data : :class:`numpy.ndarray`
            Vectorised coefficient data.

        """
        if collapse is None:
            array = self.array
        else:
            # Initialise the collapsed array.
            doublet_list = self._gen_index_list(
                self.disc.degrees, self.disc.depths, reduction=True
            )

            array = np.empty(len(doublet_list), dtype=self._dtype_collapsed)
            array['index'] = doublet_list

            for pos, ind in enumerate(array['index']):
                # Extract subarray matching the degree and depth of the
                # index doublet being considered.
                selector = np.all(
                    self.array['index'][:, [0, 2]] == ind, axis=-1
                )
                subarray = self.array['coefficient'][selector]

                # Collapse the extracted subarray.
                if collapse.lower() == 'mean':
                    collapsed_subarray = np.mean(subarray)
                elif collapse.lower() == 'qaudrature':
                    collapsed_subarray = np.mean(np.abs(subarray) ** 2)
                else:
                    raise ValueError(f"Unknown `collapse` option: {collapse}.")

                array['coefficient'][pos] = collapsed_subarray
                array['wavenumber'][pos] = self.disc.wavenumbers[tuple(ind)]

        # Sort array by the pivot order.  It is unsafe to sort by the
        # 'index' field which is multi-dimensional.  The private field
        # '_position' was initialised for this purpose.
        if pivot == 'natural':
            sort_order = ['_position', 'wavenumber']
        elif pivot == 'spectral':
            sort_order = ['wavenumber', '_position']
        else:
            raise ValueError(f"Unknown `pivot` option: {pivot}.")

        vectorised_data = np.sort(array, order=sort_order)['coefficient']

        return vectorised_data

    def _initialise_array(self):

        # Initialise the structured array in the natural order.
        array = np.empty(self.size, dtype=self._dtype)

        # Generate the list of index triplets and the dictionary of index
        # directory.
        triplet_list = self._gen_index_list(
            self.disc.degrees, self.disc.depths
        )
        triplet_directory = self._gen_index_directory(triplet_list)

        # Initialise index triplets and corresponding wavenumbers.
        array['index'] = triplet_list
        array['wavenumber'] = [
            self.disc.wavenumbers[ell, n] for ell, _, n in triplet_list
        ]
        array['_position'] = list(range(self.size))

        return array, triplet_directory

    def _find_position(self, key):

        # If accessed by integer, reinterprete negative values as reverse
        # indexing.
        if isinstance(key, int):
            position = key if key >= 0 else key % self.size
            if position > self.size:
                raise IndexingError(
                    f"Index {position} out of bound for key: {key}."
                )
            return position

        # Access by slice.
        if isinstance(key, slice):
            return key

        # If accessed by a sequence of (degree, order, depth); any list
        # is turned into a tuple.
        if isinstance(key, Sequence) and not isinstance(key, str):
            position = self._directory[tuple(key)]
            return position

        # Access by string of the form 'degree_<ell>'.
        if isinstance(key, str):
            degree = int(key.split('_')[-1])
            degree_idx = self.disc.degrees.index(degree)
            start = sum(self.disc.mode_counts[:degree_idx])
            stop = sum(self.disc.mode_counts[:(degree_idx + 1)])
            position = slice(start, stop)
            return position

        raise TypeError(f"Invalid type for key: {key}.")

    @staticmethod
    def _gen_index_list(degrees, depths, reduction=False):

        if reduction:
            index_list = [
                (ell, n + 1)
                for ell, nmax in zip(degrees, depths)
                for n in range(nmax)
            ]
        else:
            index_list = [
                (ell, m, n + 1)
                for ell, nmax in zip(degrees, depths)
                for m in range(- ell, ell+1)
                for n in range(nmax)
            ]

        return index_list

    @staticmethod
    def _gen_index_directory(index_list):

        # The directory is a mapping from an index tuple to an array position.
        index_directory = {
            tuple(tup): pos for pos, tup in enumerate(index_list)
        }

        return index_directory


class CartesianArray(DataArray):
    r"""Structured array for Cartesian decomposition of cosmological data.

    Array is initialised with three fields: the 'order' field of the
    Legendre multipoles, the 'wavenumber' field of :math:`k`-bin centres
    and the 'power' field for power spectrum multipole measurements.

    Parameters
    ----------
    orders : list of tuple of int
        Orders of the power spectrum multipole.
    wavenumbers : float, array_like
        Wavenumbers of the multipole data.
    mode_counts : list of int or None, optional
        Mode counts in wavenumber bins (default is `None`).
    shot_noise : float or None, optional
        Shot noise level (default is `None`).

    Attributes
    ----------
    array : :class:`numpy.ndarray`
        Structured NumPy array.
    size : int
        Total number of elements in the array.  This should equal the
        product of the numbers of wavenumbers and multipoles.
    attrs : dict
        Initialisation parameters as attributes.

    """

    # Class-wide ``numpy.dtype`` for the structured array.
    _dtype = np.dtype({
        'names': ['order', 'wavenumber', 'power'],
        'formats': ['i4', 'f8', 'f8'],
    })

    def __init__(self, orders, wavenumbers, mode_counts=None, shot_noise=None):

        self.size = np.size(orders) * np.size(wavenumbers)

        orders = np.sort(orders).tolist()

        wavenumbers = np.array(wavenumbers)[np.argsort(wavenumbers)]

        if mode_counts is not None:
            mode_counts = np.array(mode_counts)[np.argsort(wavenumbers)]

        self.attrs = dict(
            orders=orders,
            wavenumbers=wavenumbers,
            mode_counts=mode_counts,
            shot_noise=shot_noise,
        )

        self.array, self._directory = self._initialise_array()

    def __getitem__(self, key):
        """Access the 'power' field.

        The access key can be an integer positional index, a slice, a tuple
        of (order, wavenumber) or a string, e.g. ``[-1]``, ``[:]``,
        ``[(0, 0.04)]`` or ``'power_0'``.

        Parameters
        ----------
        key : int, slice, tuple(int, float) or str
            'power' field access key.

        Returns
        -------
        float
            'power' field data entry.

        """
        position = self._find_position(key)

        return self.array['power'][position]

    def __setitem__(self, key, value):
        """Set the 'power' field value(s).

        Parameters
        ----------
        key : int, tuple(int, float) or slice
            'power' field access key.
        value : float
            'power' field data entry.

        See Also
        --------
        :meth:`.CartesianArray.__getitem__`

        """
        position = self._find_position(key)

        self.array['power'][position] = value

    def __getstate__(self):

        return self.__dict__

    def __setstate__(self, state):

        for attr, value in state.items():
            setattr(self, attr, value)

    def vectorise(self, pivot):
        r"""Return a data vector from the 'power' field.

        Vectorisation is performed by *pivoting* in either of the following
        orders of precedence---

            * 'order': ordered by multipole order :math:`\ell`;
            * 'wavenumber': ordered by wavenumber :math:`k`.

        Parameters
        ----------
        pivot : {'order', 'wavenumber'}
            Pivot order for vectorisation.

        Returns
        -------
        vectorised_data : :class:`numpy.ndarray`
            Vectorised power spectrum data.

        """
        if pivot == 'order':
            sort_order = ['order', 'wavenumber']
        elif pivot == 'wavenumber':
            sort_order = ['wavenumber', 'order']
        else:
            raise ValueError(f"Unknown `pivot` option: {pivot}.")

        return np.sort(self.array, order=sort_order)['power']

    def _initialise_array(self):

        # Initialise the structured array ordered by order.
        array = np.empty(self.size, dtype=self._dtype)

        array['order'] = np.repeat(
            self.attrs['orders'], len(self.attrs['wavenumbers']))
        array['wavenumber'] = np.tile(
            self.attrs['wavenumbers'], len(self.attrs['orders'])
        )

        # Generate the dictionary mapping (order, wavenumber) tuples to
        # array positions.
        directory = self._gen_directory(
            self.attrs['orders'], self.attrs['wavenumbers']
        )

        return array, directory

    def _find_position(self, key):

        # If accessed by integer, reinterprete negative values as reverse
        # indexing.
        if isinstance(key, int):
            position = key if key >= 0 else key % self.size
            if position > self.size:
                raise IndexingError(
                    f"Index {position} out of bound for key: {key}."
                )
            return position

        # Access by slice.
        if isinstance(key, slice):
            return key

        # Access by sequence of (order, wavenumber).
        if isinstance(key, Sequence) and not isinstance(key, str):
            position = self._directory[tuple(key)]
            return position

        # Access by string of the form 'power_<ell>'.
        if isinstance(key, str):
            order = int(key.split('_')[-1])
            order_idx = self.attrs['orders'].index(order)
            length = len(self.attrs['wavenumbers'])
            position = slice(order_idx * length, (order_idx + 1) * length)
            return position

        raise TypeError(f"Invalid type for key: {key}.")

    @staticmethod
    def _gen_directory(orders, wavenumbers):

        directory = {
            tuple(key): pos
            for pos, key in enumerate(product(orders, wavenumbers))
        }

        return directory



import pytest

from harmonia.algorithms.discretisation import DiscreteSpectrum

FIXTURE_ARGS = dict(
    radius=100.,
    condition='dirichlet',
    highcut=0.1,
    lowcut=0.,
    maxdeg=None,
    mindeg=0,
)


@pytest.fixture(scope='module')
def discrete_spectrum():
    return DiscreteSpectrum(**FIXTURE_ARGS)


@pytest.mark.parametrize(
    "ell,n,k,kappa",
    [
        (2, 1, 0.0576346, 0.0000729768),
    ]
)
def test_DiscreteSpectrum(discrete_spectrum, ell, n, k, kappa):

    assert discrete_spectrum.degrees == [0, 1, 2, 3, 4, 5], \
        "Incorrect spherical degrees for DiscreteSpectrum test instance."
    assert discrete_spectrum.depths == [3, 2, 2, 1, 1, 1], \
        "Incorrect spherical depths for DiscreteSpectrum test instance."
    assert discrete_spectrum.mode_counts == [
        (2*ell + 1) * nmax
        for ell, nmax in zip(
            discrete_spectrum.degrees, discrete_spectrum.depths
        )
    ], "Incorrect mode counts for DiscreteSpectrum test instance."

    assert discrete_spectrum.wavenumbers[(ell, n)] == pytest.approx(k), \
        "Incorrect wavenumbers for DiscreteSpectrum test instance."
    assert discrete_spectrum.normalisations[(ell, n)] \
        == pytest.approx(kappa), \
        "Incorrect normalisations for DiscreteSpectrum test instance."



r"""
Scale dependence (:mod:`~harmonia.cosmology.scale_dependence`)
===========================================================================

Compute scale-dependence modifications to galaxy clustering from local
primordial non-Gausianity :math:`f_\textrm{NL}`.

The scale-independent linear bias is modified as

    .. math::

        b_1(z) \mapsto b_1(z) + f_\textrm{NL} [b_1(z) - p] A(k,z) \,,

where :math:`p` is a tracer-dependent parameter and the scale-dependence
modification kernel

    .. math::

        A(k,z) = 1.27 \left( \frac{H_0}{\mathrm{c}} \right)^2
            \frac{3\Omega_\textrm{m,0} \delta_\textrm{c}}{k^2 D(z) T(k)}

relates to the cosmological model and its growth factor :math:`D(z)`
(normalised to unity at the current epoch; hence the numerical factor
1.27) and transfer function :math:`T(k)` (normalised to unity as
:math:`k \to 0`).

.. autosummary::

    scale_dependence_modification
    scale_dependent_bias
    modified_power_spectrum

|

"""
# pylint: disable=no-name-in-module
from astropy.constants import c
from nbodykit import cosmology


def scale_dependence_modification(cosmo, redshift):
    r"""Return the scale-dependence modification kernel :math:`A(k,z)` for
    a given cosmological model.

    Parameters
    ----------
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        Cosmological model.
    redshift : float
        Redshift at which quantities are evaluated.

    Returns
    -------
    callable
        Scale-dependence modification kernel as a function of wavenumber
        (in :math:`h`/Mpc).

    """
    SPHERICAL_COLLAPSE_CRITICAL_OVERDENSITY = 1.686
    SPEED_OF_LIGHT_IN_KM_PER_S = c.to('km/s').value
    GROWTH_FACTOR_NORMALISATION = 1.27

    num_factors = GROWTH_FACTOR_NORMALISATION \
        * 3 * cosmo.Omega0_m * SPHERICAL_COLLAPSE_CRITICAL_OVERDENSITY \
        * (cosmo.H0 / SPEED_OF_LIGHT_IN_KM_PER_S) ** 2

    transfer_func = cosmology.power.transfers.CLASS(cosmo, redshift=redshift)

    return lambda k: num_factors / (k**2 * transfer_func(k))


def scale_dependent_bias(b_1, f_nl, cosmo, redshift=0., tracer_p=1.):
    r"""Return the scale-dependent bias modulated by local primordial
    non-Gaussianity.

    Parameters
    ----------
    b_1 : float
        Scale-independent linear bias at input redshift.
    f_nl : float
        Local primordial on-Gaussianity.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        Cosmological model.
    redshift : float, optional
        Redshift at which quantities are evaluated (default is 0.).
    tracer_p : float, optional
        Tracer-dependent parameter :math:`p` (default is 1.).

    Returns
    -------
    callable
        Scale-dependent bias as a function of wavenumber (in
        :math:`h`/Mpc).

    """
    scale_dependence = scale_dependence_modification(cosmo, redshift)

    def b_k(k):
        return b_1 + f_nl * (b_1 - tracer_p) * scale_dependence(k)

    return b_k


def modified_power_spectrum(b_1, f_nl, cosmo, redshift=0., tracer_p=1.,
                            nbar=None, contrast=None):
    r"""Return the tracer power spectrum modified by primordial
    non-Gaussianity.

    Parameters
    ----------
    b_1 : float
        Scale-independent linear bias at input redshift.
    f_nl : float
        Local primordial non-Gaussianity.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology`
        Cosmological model.
    redshift : float, optional
        Redshift at which quantities are evaluated (default is 0.).
    tracer_p : float, optional
        Tracer-dependent parameter :math:`p` (default is 1.).
    nbar : float or None, optional
        If not `None` (default), add ``1/nbar`` as shot noise to the
        resulting power spectrum.
    contrast : float or None, optional
        If not `None` (default), add additional ``1/(contrast*nbar)``
        as shot noise to the resulting power spectrum.

    Returns
    -------
    callable
        Tracer power spectrum modified by primordial non-Gaussianity as a
        function of wavenumber (in :math:`h`/Mpc).

    """
    bias = scale_dependent_bias(
        b_1, f_nl, cosmo, redshift=redshift, tracer_p=tracer_p
    )

    power_spectrum = cosmology.LinearPower(cosmo, redshift=redshift)

    alpha = 0. if contrast is None else 1 / contrast
    shot_noise = 0. if nbar is None else (1 + alpha) / nbar

    return lambda k: bias(k) ** 2 * power_spectrum(k) + shot_noise



import pytest

from harmonia.surveyor.synthesis import (
    generate_compression_matrix,
)


@pytest.mark.skip(reason="Unimplemented.")
def test_generate_compression_matrix():
    print("Unimplemented test.")
    raise NotImplementedError



r"""
Two-point correlator models (:mod:`~harmonia.reader.models`)
===========================================================================

Compute Fourier-space two-point correlator models.


Cartesian multipoles
---------------------------------------------------------------------------

The standard Kaiser model of plane-parallel power spectrum multipoles is
implemented,

.. math::

    \begin{align*}
        P_0(k, z) &=
            \left[
                b_1(z)^2 + \frac{2}{3} b(z, k) f(z) + \frac{1}{5} f(z)^2
            \right] P_\mathrm{m}(k, z) \,, \\
        P_2(k, z) &=
            \left[
                \frac{4}{3} b(z, k) f(z) + \frac{4}{7} f(z)^2
            \right] P_\mathrm{m}(k, z) \,, \\
        P_4(k, z) &= \frac{8}{35} f(z)^2 P_\mathrm{m}(k, z) \,,
    \end{align*}

with shot noise, window convolution and the integral constraint taken
into account [1]_ [2]_; here :math:`f(z)` is the linear growth rate,
:math:`b(z, k) = b_1(z) + f_\textrm{NL} \Delta b(k, z)` is the
scale-dependent linear bias including the modification :math:`\Delta b`
due to local primordial non-Gaussianity :math:`f_\textrm{NL}` (see
:mod:`~harmonia.cosmology.scale_dependence`), and :math:`P_{\textrm{m}}`
is the matter power spectrum.


.. [1] Wilson M. J. et al., 2017. MNRAS 464(3), 3121–3130.
    [arXiv: `1511.07799 <https://arxiv.org/abs/1511.07799>`_]

.. [2] Beutler F. et al., 2017. MNRAS 466(2), 2242–2260.
    [arXiv: `1607.03150 <https://arxiv.org/abs/1607.03150>`_]


Spherical correlator
---------------------------------------------------------------------------

Spherical 2-point correlators are computed from coupling coefficients
(see :mod:`~harmonia.reader.couplings`) as a sum of the signal part

.. math::

    \left\langle \delta_\mu \delta_\nu \right\rangle = \sum_\sigma
        M_{\mu\sigma} M^*_{\nu\sigma} \left[
            b_*(k_\sigma) \Phi_{\mu\sigma} + f_* \Upsilon_{\mu\sigma}
        \right] \left[
            b_*(k_\sigma) \Phi_{\nu\sigma} + f_* \Upsilon_{\nu\sigma}
        \right] \kappa_\sigma^{-1} P_{\textrm{m}*}(k_\sigma) \,,

and the shot noise part

.. math::

    \left\langle \epsilon_\mu \epsilon_\nu \right\rangle =
        \frac{1}{\bar{n}} M_{\mu\nu} \int \operatorname{d}\!r r^2
        (w^2\phi)(r) j_\mu(r) j_\nu(r) \,,

where :math:`*` denotes quantities computed at the fiducial epoch
:math:`z_*`, :math:`\kappa` denotes the normalisation coefficients (see
:mod:`~harmonia.algorithms.discretisation`), :math:`\bar{n}` is the mean
number density and
:math:`j_\mu(r) \equiv j_{\ell_\mu}(k_{\ell_\mu n_\mu} r)`.


.. autosummary::

    CartesianMultipoles
    SphericalCorrelator

|

"""
import logging
import warnings

import numpy as np
from scipy.integrate import IntegrationWarning
from scipy.interpolate import InterpolatedUnivariateSpline as Spline
from mcfit import P2xi, xi2P
from nbodykit.lab import cosmology

from harmonia.algorithms.arrays import CartesianArray, SphericalArray
from harmonia.algorithms.integration import radial_integral
from harmonia.cosmology.geometry import (
    differential_AP_distortion,
    redshift_from_distance,
)
from harmonia.cosmology.scale_dependence import scale_dependence_modification
from harmonia.utils import Progress, mpi_compute, restore_warnings

from .couplings import Couplings, SphericalCoefficientWarning, _group_couplings
from ._kernels import shot_noise_kernel


class CartesianMultipoles:
    """Power spectrum multipoles predicted for a given cosmological model
    and survey specifications.

    Parameters
    ----------
    wavenumbers : float, array_like
        Wavenumbers at which the model is evaluated.
    redshift : float
        Redshift at which the model is evaluated.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology` *or None, optional*
        Baseline cosmological model used to produce the transfer function
        and power spectrum and to compute the linear growth rate.
        This can be subsequently updated when calling
        :meth:`~.convolved_power_multipoles`.  If `None` (default) and not
        subsequently updated, primordial non-Gaussianity modifications
        cannot be computed.
    power_spectrum : callable or None, optional
        Baseline linear matter power spectrum model at `redshift`.
        Ignored when `cosmo` is provided.  This cannot be `None` (default)
        unless it is subsequently updated when calling
        :meth:`~.convolved_power_multipoles`.
    growth_rate : float or None, optional
        Baseline linear growth rate at `redshift`.  If `None` (default),
        this is set by `cosmo` if it is provided; otherwise this is set
        to 0.  This can be subsequently updated when calling
        :meth:`~.two_point_correlator` or :meth:`~.correlator_matrix`.
    mask_multipoles : :class:`numpy.ndarray` *or None, optional*
        Survey mask multipoles given at sampled separations (default
        is `None`).  Orders and sampled separations must be sorted.
    window_multipoles : :class:`~.CartesianArray` *or None, optional*
        Survey window multipoles given at sampled wavenumbers (default
        is `None`).  If provided, integral constraint corrections
        are applied.  Orders and sampled wavenumbers must be sorted.

    Attributes
    ----------
    mask_multipoles : :class:`numpy.ndarray` or None
        Survey mask multipoles given at sampled separations.
    window_multipoles : :class:`~.CartesianArray` or None
        Survey window multipoles given at sampled wavenumbers.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology` or None
        Cosmological model.
    matter_power_spectrum : callable
        Matter power spectrum at `redshift`.
    growth_rate : float
        Linear growth rate at `redshift`.
    attrs : dict
        Any other attributes inherited from input parameters.

    """

    _LOG_K_RANGE_SAMP = -5, 1
    _NUM_SAMP = pow(2, 14)
    _PREDICTED_ORDERS = [0, 2, 4]

    def __init__(self, wavenumbers, redshift, cosmo=None,
                 power_spectrum=None, growth_rate=None,
                 mask_multipoles=None, window_multipoles=None):

        self.attrs = {
            'wavenumbers': wavenumbers,
            'redshift': redshift,
        }

        self._k = wavenumbers
        self._z = redshift
        self._mask_multipoles = mask_multipoles
        self._window_multipoles = window_multipoles

        # Dense sampling wavenumbers for window convolution.
        self._k_samp = np.logspace(*self._LOG_K_RANGE_SAMP, num=self._NUM_SAMP)

        # Set a baseline model.  May be subsequently updated.  Also
        # pre-compute quantities at set wavenumbers.
        self._set_baseline_model(cosmo, power_spectrum, growth_rate)

    def convolved_power_multipoles(self, orders, b_1=None, f_nl=None,
                                   nbar=None, contrast=None, tracer_p=1.,
                                   update_model_kwargs=None):
        """Compute the convolved power spectrum multipoles.

        Parameters
        ----------
        orders : list of int
            Orders of the power spectrum multipoles.  Values only allowed
            from the set {0, 2, 4}.
        b_1 : float or None, optional
            Scale-independent linear bias at input redshift.  If `None`
            (default), no tracer bias is assumed relative to the matter
            power spectrum.
        f_nl : float or None, optional
            Local primordial non-Gaussianity (default is `None`).
        nbar : float or None, optional
            Mean particle number density (in cubic :math:`h`/Mpc).  If
            `None` (default), shot noise is neglected.
        contrast : float or None, optional
            If not `None` (default), this adds additional shot noise
            ``1 / (contrast * nbar)`` from a FKP-style random catalogue.
        tracer_p : float, optional
            Tracer-dependent parameter for bias modulation by `f_nl`
            (default is 1.).
        update_model_kwargs : dict or None, optional
            Parameters `cosmo`, `power_spectrum` and `growth_rate` passed
            as keyword arguments to update the baseline cosmological model.

        Returns
        -------
        convolved_power : :class:`~.algorithms.arrays.CartesianArray`
            Convolved power spectrum multipoles.

        """
        # Updated initial baseline cosmological model if relevant
        # arguments specified.
        if update_model_kwargs is not None:
            self._set_baseline_model(**update_model_kwargs)

        b_1 = b_1 or 1.

        if f_nl is not None and self.cosmo is None:
            raise TypeError(
                "Cannot accept `f_nl` values without input `cosmo`."
            )

        alpha = 0. if contrast is None else 1 / contrast
        shot_noise = (1 + alpha) / nbar

        convolved_power = CartesianArray(
            orders, self._k, shot_noise=shot_noise
        )

        # If no survey mask, no convolution is needed.
        if self._mask_multipoles is None:
            b_k = b_1 * np.ones_like(self._k)
            if f_nl:
                b_k += self._mode_modification * f_nl * (b_1 - tracer_p)

            convolved_power[:] = np.concatenate([
                self.kaiser_factors(ell, b_k, self.growth_rate)
                * self._mode_powers
                + np.equal(ell, 0) * shot_noise
                for ell in orders
            ])

            return convolved_power

        # Get mask multipoles as a directory.
        s_mask = np.unique(self._mask_multipoles['separation'])
        mask = {
            ell: self._mask_multipoles['correlation'][
                self._mask_multipoles['order'] == ell
            ]
            for ell in np.unique(self._mask_multipoles['order'])
        }

        # Get power multipoles at densely sampled points for interpolation.
        b_k_samp = b_1 * np.ones_like(self._k_samp)
        if f_nl:
            b_k_samp += self._mode_modification_samp * f_nl * (b_1 - tracer_p)

        power_samp = {
            ell: self.kaiser_factors(ell, b_k_samp, self.growth_rate) \
                * self.matter_power_spectrum(self._k_samp)
            for ell in self._PREDICTED_ORDERS
        }

        # Hankel transform to correlation multipoles.
        with warnings.catch_warnings():  # suppress `mcfit` warning
            warnings.filterwarnings(
                'ignore',
                message=(
                    "The default value of extrap has been changed to False, "
                    "set it to True if you cannot reproduce previous results"
                )
            )
            correlation = {
                ell: Spline(
                    *P2xi(self._k_samp, l=ell, lowring=True)(
                        power_samp[ell], extrap=False
                    ), k=1
                )(s_mask)
                for ell in self._PREDICTED_ORDERS
            }

        # Convolve correlation multipoles.
        correlation_convolved = {
            ell: self._convolve(ell, correlation, mask) for ell in orders
        }

        # Hankel transform back to power multipoles.
        with warnings.catch_warnings():  # suppress warnings from `mcfit`
            warnings.filterwarnings(
                'ignore',
                message=(
                    "The default value of extrap has been changed to False, "
                    "set it to True if you cannot reproduce previous results"
                )
            )
            power_conv = {
                ell: Spline(
                    *xi2P(s_mask, l=ell, lowring=True)(
                        correlation_convolved[ell], extrap=False
                    ), k=1
                )(self._k) + np.equal(ell, 0) * shot_noise
                for ell in orders
            }

        # Apply integral constraint (ic) corrections.
        if self._window_multipoles is None:
            warnings.warn(
                "`window_multipoles` attribute is missing. "
                "Integral constraint is not corrected."
            )
        else:
            # Get window multipoles.
            k_win = np.unique(
                self._window_multipoles.array['wavenumber']
            )
            window = {
                ell: self._window_multipoles.array['power'][
                    self._window_multipoles.array['order'] == ell
                ]
                for ell in np.unique(self._window_multipoles.array['order'])
            }

            # Interpolate window multipoles at model wavenumbers.
            window_intrpl = {
                ell: Spline(k_win, window[ell])(self._k)
                for ell in power_conv
            }

            with warnings.catch_warnings():  # suppress warnings from `mcfit`
                warnings.filterwarnings(
                    'ignore',
                    message=(
                        "The default value of extrap has been changed "
                        "to False, set it to True if you cannot "
                        "reproduce previous results"
                    )
                )
                ic_correction = Spline(
                    *xi2P(s_mask, l=0, lowring=True)(
                        correlation_convolved[0], extrap=False
                    ), k=1
                )(0)

            power_conv.update({
                ell: power_conv[ell] - ic_correction * window_intrpl[ell]
                for ell in orders
            })

        convolved_power[:] = np.concatenate(
            [power_conv[ell] for ell in orders]
        )

        return convolved_power

    @staticmethod
    def kaiser_factors(ell, b, f):
        """Return the standard Kaiser power spectrum multipole as a
        multipole of the matter power spectrum.

        Notes
        -----
        Linear bias `b` and growth rate `f` must be specified at the
        same redshift.

        Parameters
        ----------
        ell : int
            Order of the multipole in {0, 2, 4}.
        b : float, array_like
            Linear bias of the tracer particles.
        f : float or None
            Linear growth rate.

        Returns
        -------
        float, array_like
            Standard Kaiser factor to multiply by the matter power
            spectrum.

        """
        if ell == 0:
            factor = b**2 + 2./3. * b * f + 1./5. * f**2
        elif ell == 2:
            factor = 4./3. * b * f + 4./7. * f**2
        elif ell == 4:
            factor = 8./35. * f**2
        else:
            raise ValueError("Multipole order must be one of {0, 2, 4}.")

        return factor

    def _set_baseline_model(self, cosmo=None, power_spectrum=None,
                            growth_rate=None):

        # If `cosmo` is not None, `self._kernel` as A(k) (see docstring),
        # `self._mode_modification` is its evaluation at wavenumbers.
        # `self._mode_powers` is the evaluated matter power spectrum.
        self.cosmo = cosmo
        if self.cosmo is None:
            self.matter_power_spectrum = power_spectrum
            self.growth_rate = growth_rate or 0.
            self._kernel = None
            self._mode_modification = None
            self._mode_modification_samp = None
        else:
            self.matter_power_spectrum = cosmology.LinearPower(
                self.cosmo, redshift=self._z, transfer='CLASS'
            )
            self.growth_rate = growth_rate if growth_rate is not None \
                else self.cosmo.scale_independent_growth_rate(self._z)
            self._kernel = scale_dependence_modification(self.cosmo, self._z)
            self._mode_modification = self._kernel(self._k)
            self._mode_modification_samp = self._kernel(self._k_samp)

        # (Re-)set pre-computed quantities.
        try:
            self._mode_powers = self.matter_power_spectrum(self._k)
            self._mode_powers_samp = self.matter_power_spectrum(self._k_samp)
        except TypeError:
            warnings.warn(
                "Cannot evaluate :attr:`matter_power_spectrum`. "
                "No baseline model is current set."
            )

    @staticmethod
    def _convolve(order, correlation, mask):

        if order == 0:
            return correlation[0] * mask[0] \
                + 1./5. * correlation[2] * mask[2] \
                + 1./9. * correlation[4] * mask[4]

        if order == 2:
            return correlation[0] * mask[2] \
                + correlation[2] * (
                    mask[0] + 2./7. * mask[2] + 2./7. * mask[4]
                ) \
                + correlation[4] * (
                    2./7. * mask[2] + 100./693. * mask[4] + 25./143. * mask[6]
                )

        if order == 4:
            return correlation[0] * mask[4] \
                + correlation[2] * (
                    18./35. * mask[2] + 20./77. * mask[4] + 45./143. * mask[6]
                ) \
                + correlation[4] * (
                    mask[0]
                    + 20./77. * mask[2] + 162./1001. * mask[4]
                    + 20./143. * mask[6] + 490./2431. * mask[8]
                )

        raise ValueError(f"Order not predicted by the model: {order}.")


class SphericalCorrelator:
    """Spherical two-point correlator predicted for a given cosmological
    model and survey specifications.

    Notes
    -----
    Summation is performed over spherical modes to compute the correlator
    model.  Owing to mode couplings, sums need to be truncated at higher
    mode wavenumbers than wavenumbers under consideration to ensure
    convergence.  We suggest the user trial discrete spectra with different
    upper wavenumber cutoffs for the couplings and use
    :meth:`~.radialised_power` to gauge convergence.  In the future,
    an automated facility for this purpose may be provided.

    Parameters
    ----------
    disc : :class:`~harmonia.algorithms.discretisation.DiscreteSpectrum`
        Discrete spectrum associated with the correlator model.
    redshift : float
        Redshift at which the model is evaluated.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology` *or None, optional*
        Baseline cosmological model used to produce the transfer function
        (and the power spectrum and linear growth rate if these are not
        set but required in model evaluation).  This can be subsequently
        updated when calling :meth:`~.two_point_correlator`
        or :meth:`~.correlator_matrix`.  If `None` (default) and not
        subsequently updated, primordial non-Gaussianity modifications
        cannot be computed.
    power_spectrum : callable or None, optional
        Baseline linear matter power spectrum model at `redshift`.
        Ignored when `cosmo` is provided; otherwise this cannot be `None`
        (default) unless it is subsequently updated when calling
        :meth:`~.two_point_correlator` or :meth:`~.correlator_matrix`.
    growth_rate : float or None, optional
        Baseline linear growth rate at `redshift`.  If `None` (default),
        this is set by `cosmo` (if provided); otherwise this is set
        to zero.  This can be subsequently updated when calling
        :meth:`~.two_point_correlator` or :meth:`~.correlator_matrix`.
    couplings : :class:`~.couplings.Couplings` *or None, optional*
        Baseline coupling coefficients consistent with the underlying
        cosmological model `cosmo`.  If `None` (default), this is compiled
        from `survey_specs` and `cosmo_specs` if either is provided;
        otherwise all couplings are assumed to be trivial (i.e. angular
        and radial couplings are Kronecker deltas).  This can be
        subsequently updated when calling :meth:`~.two_point_correlator`
        or :meth:`~.correlator_matrix`.
    coupling_disc : :class:`~.couplings.Couplings` *or None, optional*
        Discrete spectrum for the baseline coupling coefficients
        (default is `None`).  This must be provided if `couplings` is
        `None` and needs to be compiled.  To ensure sum convergence in the
        correlator model, the upper wavenumber cutoff for this spectrum
        may need to be higher than that of the correlator model,
        i.e. `disc`.
    survey_specs : dict{str: callable or None} or None, optional
        Survey specification functions to be passed as `survey_specs` to
        :class:`~harmonia.reader.couplings.Couplings` when couplings are
        compiled.  Also used in shot noise calculations.
    cosmo_specs : dict{str: callable, bool or None} or None, optional
        Baseline cosmological specification functions to be passed as
        `cosmo_specs` to :class:`~harmonia.reader.couplings.Couplings`
        when couplings are compiled.  If not `None` (default), it must be
        a dictionary holding keys listed in :class:`~.couplings.Couplings`:
        if callable values are passed to the keys ``'chi_of_z'``,
        ``'clustering_evolution'``, ``'growth_evolution'`` or
        ``'differential_AP_distortion'``, they should be consistent with
        the current :attr:`cosmo`; otherwise `True` can be passed here and
        their values are then derived from `cosmo` (which must then be
        set), or unspecified keys assume `None` values; also note some
        keys are linked and values must be simultaneously provided.  This
        can be subsequently updated when calling
        :meth:`~.two_point_correlator` or :meth:`~.correlator_matrix`.
    ini_shot_noise : bool, optional
        If `True` (default), shot noise integrals are evaluated upon
        initialisation.
    comm : :class:`mpi4py.MPI.Comm` *or None, optional*
        MPI communicator.  If `None` (default), no multiprocessing
        is invoked.

    Attributes
    ----------
    couplings : :class:`nbodykit.cosmology.cosmology.Cosmology` or None
        Spherical coupling coefficients.
    cosmo : :class:`nbodykit.cosmology.cosmology.Cosmology` or None
        Cosmological model.
    matter_power_spectrum : callable
        Matter power spectrum at `redshift`.
    growth_rate : float
        Linear growth rate at `redshift`.
    attrs : dict
        Any other attributes inherited from input parameters.

    See Also
    --------
    :class:`~harmonia.reader.couplings.Couplings`
        More details related to `couplings` and especially the
        `cosmo_specs` parameter.

    """

    def __init__(self, disc, redshift, cosmo=None, power_spectrum=None,
                 growth_rate=None, couplings=None, coupling_disc=None,
                 survey_specs=None, cosmo_specs=None, ini_shot_noise=True,
                 comm=None):

        self.comm = comm
        self.logger = logging.getLogger(self.__class__.__name__)

        self.attrs = {
            'disc': disc,
            'redshift': redshift,
        }

        self._disc = disc
        self._z = redshift

        # NOTE: This assignment is needed to help :meth:`_render_cosmo_specs`
        # set the couplings next, which are not initialised with `cosmo`,
        # whose formal assignment must come after in
        # :meth:`_set_baseline_model`, as therein the couplings are processed.
        self.cosmo = cosmo

        self._survey_specs = survey_specs
        self._cosmo_specs = self._render_cosmo_specs(cosmo_specs)

        if couplings is None:
            self._dummy_disc = coupling_disc
            self.couplings = Couplings(
                self._dummy_disc,
                survey_specs=self._survey_specs,
                cosmo_specs=self._cosmo_specs,
                initialise=True, comm=self.comm
            )
        else:
            self._dummy_disc = couplings.disc
            self.couplings = couplings

        # Group couplings for mass access.
        self._grouped_couplings = _group_couplings(self.couplings)

        # Set a baseline model and pre-compute power and transfer function for
        # fixed modes.  May be subsequently updated.
        # NOTE: This step must come after setting the `couplings` attribute,
        # since it also (re-)processes the radial/RSD couplings.
        self._set_baseline_model(
            cosmo=cosmo, power_spectrum=power_spectrum, growth_rate=growth_rate
        )

        # Compile shot noise levels and cosmology-independent angular
        # coupling sums as directories to reduce computational redundancy.
        if ini_shot_noise:
            self._shot_noise_levels = self._compile_shot_noise_levels()
        else:
            self._shot_noise_levels = None
        if self.couplings is None:
            self._angular_sums = None
        else:
            self._angular_sums = self._compile_angular_sums()

    def two_point_correlator(self, mu, nu, b_1, f_nl=None, nbar=None,
                             contrast=None, tracer_p=1.,
                             update_model_kwargs=None):
        """Compute two-point correlator for given triplet indices.

        Parameters
        ----------
        mu, nu : tuple(int, int, int)
            Coefficient triplet index.
        b_1 : float
            Scale-independent linear bias at input redshift.
        f_nl : float or None, optional
            Local primordial non-Gaussianity (default is `None`).
        nbar : float or None, optional
            Mean particle number density (in cubic :math:`h`/Mpc).  If
            `None` (default), shot noise is neglected.
        contrast : float or None, optional
            If not `None` (default), this adds additional shot noise
            ``1 / (contrast * nbar)`` from a FKP-style random catalogue.
        tracer_p : float, optional
            Tracer-dependent parameter for bias modulation by `f_nl`
            (default is 1.).
        update_model_kwargs : dict or None, optional
            Parameters `cosmo`, `power_spectrum`, `growth_rate` and
            `cosmo_specs` passed as keyword arguments to update the
            baseline cosmological model.  If `cosmo_specs` is passed
            (including `None` values), radial and RSD couplings will be
            updated.

        Returns
        -------
        complex
            Spherical Fourier 2-point function value for given triplet
            indices.  If `cosmo_specs` is passed, radial and RSD couplings
            will be updated.

        """
        if update_model_kwargs is not None:
            if 'cosmo_specs' in update_model_kwargs:
                update_model_kwargs.update({'update_couplings': True})
            self._set_baseline_model(**update_model_kwargs)

        f = self.growth_rate
        # Grand sum over dummy degree index, say `ell_sigma`.
        correlator_value = 0.
        for ell_idx, ell_sigma in enumerate(self._dummy_disc.degrees):
            # Fetch angular sum over order `m_sigma`.
            try:
                angular_sum = self._get_angular_sums(mu, nu)[ell_sigma]
            except TypeError:
                # Non-subscriptable 0.'s and 1.'s independent of degree.
                angular_sum = self._get_angular_sums(mu, nu)

            # Perform radial sum over depth `n_sigma`.
            b_k = b_1
            if f_nl:
                b_k += f_nl  * (b_1 - tracer_p) \
                    * self._mode_modification[ell_idx]
            if f:
                radial_sum = np.sum(
                    (
                        b_k
                        * self._get_grouped_couplings('radial', mu)[ell_idx]
                        + f
                        * self._get_grouped_couplings('rsd', mu)[ell_idx]
                    ) * (
                        b_k
                        * self._get_grouped_couplings('radial', nu)[ell_idx]
                        + f
                        * self._get_grouped_couplings('rsd', nu)[ell_idx]
                    ) * self._normalised_mode_powers[ell_idx]
                )
            else:
                radial_sum = np.sum(
                    b_k * self._get_grouped_couplings('radial', mu)[ell_idx]
                    * b_k * self._get_grouped_couplings('radial', nu)[ell_idx]
                    * self._normalised_mode_powers[ell_idx]
                )

            correlator_value += angular_sum * radial_sum

        if nbar is not None:
            alpha = 0. if contrast is None else 1 / contrast
            correlator_value += (1 + alpha) / nbar \
                * self._get_shot_noise_level(mu, nu)

        return correlator_value

    def correlator_matrix(self, pivot, b_1=None, f_nl=None, nbar=None,
                          contrast=None, tracer_p=1., diagonal=False,
                          shot_noise_only=False, update_model_kwargs=None,
                          report_progress=False):
        """Compute two-point correlator matrix for some vetorisation of all
        spectrum modes.

        Parameters
        ----------
        pivot : {'natural', 'spectral'}
            Pivot order for vectorisation.
        b_1 : float
            Scale-independent linear bias at input redshift.
        f_nl : float or None, optional
            Local primordial non-Gaussianity (default is `None`).
        nbar : float or None, optional
            Mean particle number density (in cubic :math:`h`/Mpc).  If
            `None` (default), shot noise is neglected.
        contrast : float or None, optional
            If not `None` (default), this adds additional shot noise
            ``1 / (contrast * nbar)`` from a FKP-style random catalogue.
        tracer_p : float, optional
            Tracer-dependent parameter for bias modulation by `f_nl`
            (default is 1.).
        diagonal : bool, optional
            If `True` (default is `False`), return only the diagonal
            matrix part.
        shot_noise_only : bool, optional
            If `True` (default is `False`), return only the shot noise
            correlator matrix.
        update_model_kwargs : dict or None, optional
            Parameters `cosmo`, `power_spectrum`, `growth_rate` and
            `cosmo_specs` passed as keyword arguments to update the
            baseline cosmological model.  If `cosmo_specs` is passed
            (including `None` values), radial and RSD couplings will be
            updated.
        report_progress : bool, optional
            If `True` (default is `False`), progress status will be
            reported.

        Returns
        -------
        complex :class:`numpy.ndarray`
            Two-point correlator matrix vectorised by the given pivot.

        See Also
        --------
        :class:`~harmonia.algorithms.arrays.SphericalArray`

        """
        if update_model_kwargs is not None:
            if 'cosmo_specs' in update_model_kwargs:
                update_model_kwargs.update({'update_couplings': True})
            self._set_baseline_model(**update_model_kwargs)

        index_array = SphericalArray(self._disc)
        # NOTE: Careful sorting is needed here.
        if pivot == 'natural':
            index_vector = [
                tuple(index)
                for index in index_array.array['index']
            ]
        elif pivot == 'spectral':
            index_vector = [
                tuple(index)
                for index in index_array.array['index'][
                    np.argsort(index_array.array['wavenumber'])
                ]
            ]
        else:
            raise ValueError(f"Invalid pivot option: {pivot}.")

        dim = len(index_vector)

        ## Diagonal: Only return diagonal matrix.

        if diagonal:
            # Shot noise contribution.
            if nbar is None:
                shot_noise_diag = np.zeros(dim, dtype=complex)
            else:
                alpha = 0. if contrast is None else 1 / contrast
                shot_noise_diag = (1 + alpha) / nbar * np.asarray([
                    self._get_shot_noise_level(mu, mu)
                    for mu in index_vector
                ])

            if shot_noise_only:
                return np.diag(shot_noise_diag)

            # Signal contribution.
            signal_diag = np.asarray([
                self.two_point_correlator(
                    mu, mu, b_1=b_1, f_nl=f_nl, tracer_p=tracer_p
                )
                for mu in index_vector
            ])

            two_point_corr_diag = signal_diag + shot_noise_diag

            return np.diag(two_point_corr_diag)

        ## Full: Return the entire correlator matrix.

        # Shot noise contribution.
        shot_noise_mat = np.zeros((dim,) * 2, dtype=complex)

        if nbar is not None:
            alpha = 0. if contrast is None else 1 / contrast
            shot_noise_amplitude = (1 + alpha) / nbar
            for row_idx in range(dim):
                for col_idx in range(row_idx + 1):
                    mu, nu = index_vector[row_idx], index_vector[col_idx]
                    shot_noise_mat[row_idx, col_idx] = shot_noise_amplitude * \
                        self._get_shot_noise_level(mu, nu)

        if shot_noise_only:
            return shot_noise_mat

        # Signal contribution.
        if report_progress:
            progress = Progress(
                (1 + dim) * dim / 2,
                process_name="correlator matrix",
                logger=self.logger, comm=self.comm
            )
            idx = 0
        else:
            progress = None

        signal_mat = np.zeros((dim,) * 2, dtype=complex)
        for row_idx in range(dim):
            for col_idx in range(row_idx + 1):
                mu, nu = index_vector[row_idx], index_vector[col_idx]
                signal_mat[row_idx, col_idx] = self.two_point_correlator(
                    mu, nu, b_1=b_1, f_nl=f_nl, tracer_p=tracer_p
                )
                # pylint: disable=multiple-statements
                if progress:
                    progress.report(idx); idx += 1

        # Use Hermitian property to fill in the strictly upper triangular part.
        two_point_corr_mat = signal_mat + shot_noise_mat
        two_point_corr_mat[np.triu_indices(dim, k=1)] = np.conj(
            np.transpose(two_point_corr_mat)
        )[np.triu_indices(dim, k=1)]

        return two_point_corr_mat

    def radialised_power(self, b_1=None, f_nl=None, nbar=None, contrast=None,
                         tracer_p=1., shot_noise_only=False):
        """Compute the radialised spherical mode power.

        Notes
        -----
        This relies on :meth:`~.correlator_matrix` with ``diagonal=True``.
        Results are only meaningful in the radialisation limit when
        all coupling coefficients are trivial (i.e. Kronecker deltas) so
        that spherical modes are mutually independent [3]_.  Mode power is
        averaged over equivalent spherical orders and suitably normalised
        so that it matches the Cartesian power spectrum for a full-sky
        statistically isotripoc map.

        Parameters
        ----------
        b_1 : float
            Scale-independent linear bias at input redshift.
        f_nl : float or None, optional
            Local primordial non-Gaussianity (default is `None`).
        nbar : float or None, optional
            Mean particle number density (in cubic :math:`h`/Mpc).  If
            `None` (default), shot noise is neglected.
        contrast : float or None, optional
            If not `None` (default), this adds additional shot noise
            ``1 / (contrast * nbar)`` from a FKP-style random catalogue.
        tracer_p : float, optional
            Tracer-dependent parameter for bias modulation by `f_nl`
            (default is 1.).
        shot_noise_only : bool, optional
            If `True` (default is `False`), return only the shot noise
            power.

        Returns
        -------
        dict
            Radialised spherical mode powers at wavenumbers with mode
            indices.

        See Also
        --------
        :attr:`~harmonia.mapper.map_transform.SphericalMap.mode_power`


        .. [3] Rassat A. & Refregier A., 2012. A&A 540, A115.
           [arXiv: `1112.3100 <https://arxiv.org/abs/1112.3100>`_]

        """
        diagonal_correlator = self.correlator_matrix(
            'natural', b_1=b_1, f_nl=f_nl, nbar=nbar, contrast=contrast,
            tracer_p=tracer_p, diagonal=True, shot_noise_only=shot_noise_only
        )

        index_vector = self._gen_operable_indices(self._disc)

        normalisation_vector = [
            self._disc.normalisations[index[0], index[-1]]
            for index in index_vector
        ]

        mode_power_array = SphericalArray(self._disc)

        mode_power_array[:] = np.multiply(
            normalisation_vector, np.diag(diagonal_correlator)
        )

        mode_powers = mode_power_array.vectorise('spectral', collapse='mean')
        wavenumbers = np.sort(list(self._disc.wavenumbers.values()))

        return {
            'wavenumbers': wavenumbers,
            'mode_powers': np.real(mode_powers),
        }

    def _set_baseline_model(self, **kwargs):
        """(Re-)set baseline model.

        Parameters
        ----------
        **kwargs
            Parameters `cosmo`, `power_spectrum`, `growth_rate`,
            `cosmo_specs` and `update_couplings` to be set or reset.

        """
        ## `self._kernel` is A(k) (see docstring), `self._mode_modification`
        ## is its evaluation at wavenumbers, and `self._normalised_mode_powers`
        ## is the evaluated matter power spectrum normalised 'spherically'.

        if 'cosmo' in kwargs:
            self.cosmo = kwargs['cosmo']
            # (Re-)set pre-computed quantities.
            if self.cosmo is None:
                self._kernel = None
                self._mode_modification = None
            else:
                self._kernel = scale_dependence_modification(
                    self.cosmo, self._z
                )
                self._mode_modification = [
                    np.asarray([
                        self._kernel(
                            self._dummy_disc.wavenumbers[ell_sigma, n_sigma]
                        )
                        for n_sigma in range(1, nmax_sigma + 1)
                    ])
                    for ell_sigma, nmax_sigma in zip(
                        self._dummy_disc.degrees, self._dummy_disc.depths
                    )
                ]

        if 'power_spectrum' in kwargs:
            if self.cosmo is None:
                self.matter_power_spectrum = kwargs['power_spectrum']
            else:
                self.matter_power_spectrum = cosmology.LinearPower(
                    self.cosmo, redshift=self._z, transfer='CLASS'
                )
            # (Re-)set pre-computed quantities.
            try:
                self._normalised_mode_powers = [
                    np.asarray([
                        self.matter_power_spectrum(
                            self._dummy_disc.wavenumbers[ell_sigma, n_sigma]
                        ) / self._dummy_disc.normalisations[ell_sigma, n_sigma]
                        for n_sigma in range(1, nmax_sigma + 1)
                    ])
                    for ell_sigma, nmax_sigma in zip(
                        self._dummy_disc.degrees, self._dummy_disc.depths
                    )
                ]
            except TypeError:
                warnings.warn(
                    "Cannot evaluate :attr:`matter_power_spectrum`. "
                    "No baseline model is current set."
                )

        if 'growth_rate' in kwargs:
            if self.cosmo is None:
                self.growth_rate = kwargs['growth_rate'] or 0.
            else:
                self.growth_rate = kwargs['growth_rate'] \
                    if kwargs['growth_rate'] is not None \
                    else self.cosmo.scale_independent_growth_rate(self._z)

        ## Only necessary to update radial and RSD couplings (based on the
        ## latest `cosmo` attribute) when `cosmo_specs` is specified.  Any
        ## previously computed angular couplings are attached.

        if kwargs.get('update_couplings'):
            try:
                self._cosmo_specs = \
                    self._render_cosmo_specs(kwargs['cosmo_specs'])
            except KeyError:
                warnings.warn(
                    "`update_couplings` is ignored as `cosmo_specs` was not "
                    "passed. If you intend it to be `None` it still needs to "
                    "be explicitly passed."
                )
            else:
                try:
                    current_angular_couplings = \
                        self.couplings.couplings['angular']
                except (AttributeError, KeyError):
                    current_angular_couplings = None
                self.couplings = Couplings(
                    self._dummy_disc,
                    survey_specs=self._survey_specs,
                    cosmo_specs=self._cosmo_specs,
                    external_angular_couplings=current_angular_couplings,
                    initialise=True, comm=self.comm
                )
                self._grouped_couplings = _group_couplings(self.couplings)

    def _render_cosmo_specs(self, cosmo_specs):

        # This renders `cosmo_specs` ready to be passed to
        # :class:`~.couplings.Couplings` for coupling compilation.

        _error_msg = (
            "Cannot convert '{}' from True to callable as "
            "the `cosmo` attribute is not set."
        )

        if cosmo_specs is None:
            return None

        if isinstance(cosmo_specs, dict):
            # pylint: disable=protected-access
            _cosmo_specs = dict.fromkeys(Couplings._cosmo_specs.keys())
            if not set(cosmo_specs.keys()).issubset(_cosmo_specs.keys()):
                raise ValueError(
                    "Invalid `cosmo_specs` keys: {}.".format(
                        set(cosmo_specs.keys()) - set(_cosmo_specs.keys())
                    )
                )
            _cosmo_specs.update(cosmo_specs)

            # Convert `True` values using the `cosmo` attribute.
            if _cosmo_specs['z_from_r'] is True:
                try:
                    z_from_r = redshift_from_distance(self.cosmo)
                except AttributeError:
                    raise ValueError(_error_msg.format('z_from_r'))
                else:
                    _cosmo_specs.update({'z_from_r': z_from_r})

            if _cosmo_specs['clustering_evolution'] is True:
                try:
                    clustering_evolution = lambda z: \
                        self.cosmo.scale_independent_growth_factor(z) \
                        / self.cosmo.scale_independent_growth_factor(self._z)
                except AttributeError:
                    raise ValueError(_error_msg.format('clustering_evolution'))
                else:
                    _cosmo_specs.update({
                        'clustering_evolution': clustering_evolution
                    })

            if _cosmo_specs['growth_evolution'] is True:
                try:
                    growth_evolution = lambda z: \
                        self.cosmo.scale_independent_growth_rate(z) \
                        / self.cosmo.scale_independent_growth_rate(self._z)
                except AttributeError:
                    raise ValueError(_error_msg.format('growth_evolution'))
                else:
                    _cosmo_specs.update({'growth_evolution': growth_evolution})

            if _cosmo_specs['differential_AP_distortion'] is True:
                try:
                    diff_AP = differential_AP_distortion(
                        _cosmo_specs['chi_of_z'], self.cosmo.comoving_distance
                    )
                except AttributeError:
                    raise ValueError(
                        _error_msg.format('differential_AP_distortion')
                    )
                else:
                    _cosmo_specs.update({
                        'differential_AP_distortion': diff_AP
                    })

            return _cosmo_specs

        raise TypeError("Invalid `cosmo_specs` passed.")

    def _get_grouped_couplings(self, coupling_type, mu):

        return self._grouped_couplings[coupling_type][mu[0], mu[-1]]

    def _get_angular_sums(self, mu, nu):

        try:
            return self._angular_sums[(mu[0], mu[1]), (nu[0], nu[1])]
        except (KeyError, TypeError):
            try:  # use Hermitian conjugate if unavailable
                return self._angular_sums[(nu[0], nu[1]), (mu[0], mu[1])]
            except TypeError:  # non-existent, reduce to Kronecker delta
                return float(mu[0] == nu[0] and mu[1] == nu[1])

    def _get_shot_noise_level(self, mu, nu):

        try:
            return self._shot_noise_levels[mu, nu]
        except TypeError:  # if initialised with ``ini_shot_noise=False``
            self._shot_noise_levels = self._compile_shot_noise_levels()
            return self._shot_noise_levels[mu, nu]
        except KeyError:  # use Hermitian conjugate if unavailable
            return self._shot_noise_levels[nu, mu]

    def _compile_angular_sums(self):

        # Compiles all index pairs of the form (\ell_\mu, m_\mu, \ell_\nu,
        # m_\nu) where \nu is a higher index than \mu (the other cases are
        # obtained by Hermitian conjuate).
        index_pair_vector = list(
            (first_index, second_index)
            for first_index in self._gen_operable_indices(
                self._disc, subtype='angular'
            )
            for second_index in self._gen_operable_indices(
                self._disc, subtype='angular', above_from=first_index
            )
        )

        if self.comm is None or self.comm.rank == 0:
            self.logger.info("Compiling fixed angular sums...")

        # Compile, for each index pair of the form above, all its angular sums
        # indexed by a new degree index (say, `ell_sigma`) of the form:
        # \sum_{m_\ell} M_{\ell_\mu, m_\mu, \ell_\sigma, m_\sigma)
        # * M_{\ell_\nu, m_\nu, \ell_\sigma, m_\sigma)^*.
        angular_sums_by_index_pair = mpi_compute(
            index_pair_vector, self._compile_angular_sums_by_index_pair,
            comm=self.comm, process_name="angular sum compilation"
        )

        # Store compiled results in a directory, where each key is an index
        # pair of the form above, and each entry is a dictionary of angular
        # sums indexed by the new degree index (i.e. `ell_sigma`).
        angular_sums = {}
        for index_pair, angular_sums_for_index_pair \
                in zip(index_pair_vector, angular_sums_by_index_pair):
            angular_sums.update({
                tuple(index_pair): angular_sums_for_index_pair
            })

        if self.comm is None or self.comm.rank == 0:
            self.logger.info("... compiled fixed angular sums.")

        return angular_sums

    def _compile_shot_noise_levels(self):

        # Compiles all index pairs (`mu`, `nu`) where `nu` is a higher index
        # than `mu` (the other cases are obtained by Hermitian conjuate).
        index_pair_vector = list(
            (first_index, second_index)
            for first_index in self._gen_operable_indices(self._disc)
            for second_index in \
                self._gen_operable_indices(self._disc, above_from=first_index)
        )

        if self.comm is None or self.comm.rank == 0:
            self.logger.info("Compiling shot noise levels...")

        # Compile, for each index pair of the form above, the shot noise level
        # of the form: M_{\mu, \nu} * shot noise integral[selection, weight].
        shot_noise_levels_by_index_pair = mpi_compute(
            index_pair_vector, self._compile_shot_noise_levels_by_index_pair,
            comm=self.comm, process_name="shot noise level compilation"
        )

        # Store compiled results in a directory, where each key is an index
        # pair of the form above, and each entry is a complex shot noise
        # integral value including the mask.
        shot_noise_levels = {}
        with warnings.catch_warnings(record=True) as any_warnings:
            for index_pair, shot_noise_levels_for_index_pair \
                    in zip(index_pair_vector, shot_noise_levels_by_index_pair):
                shot_noise_levels.update({
                    tuple(index_pair): shot_noise_levels_for_index_pair
                })
        if any_warnings:
            restore_warnings(any_warnings)
            warnings.warn(
                "Poor shot noise integration convergence.",
                IntegrationWarning
            )

        if self.comm is None or self.comm.rank == 0:
            self.logger.info("... compiled shot noise levels.")

        return shot_noise_levels

    def _compile_angular_sums_by_index_pair(self, index_pair):

        mu, nu = index_pair

        # For each degree index (say `ell_sigma`), return \sum_{m_\sigma}
        # M_{\mu, \sigma} * M_{\nu, \sigma}^*.
        angular_sums_for_index_pair_by_degree = {
            ell_sigma: sum([
                self.couplings['angular', mu, (ell_sigma, m_sigma)]
                * np.conj(self.couplings['angular', nu, (ell_sigma, m_sigma)])
                for m_sigma in range(- ell_sigma, ell_sigma + 1)
            ])
            for ell_sigma in self._dummy_disc.degrees
        }

        return angular_sums_for_index_pair_by_degree

    def _compile_shot_noise_levels_by_index_pair(self, index_pair):

        mu, nu = index_pair

        M_mu_nu = self.couplings['angular', mu, nu]

        k_mu = self._disc.wavenumbers[mu[0], mu[-1]]
        k_nu = self._disc.wavenumbers[nu[0], nu[-1]]

        # Return the radial integral over selection(r) * weight(r)^2 *
        # j_{\ell_mu}(k_\mu r) * j_{\ell_\nu}(k_\nu r).
        with warnings.catch_warnings(record=True) as any_warning:
            shot_noise_integral = radial_integral(
                lambda r: shot_noise_kernel(
                    r, mu, nu, k_mu, k_nu,
                    selection=getattr(self._survey_specs, 'selection', None),
                    weight=getattr(self._survey_specs, 'weight', None)
                ), self._disc.attrs['boundary_radius']
            )
        if any_warning and not np.isclose(shot_noise_integral, 0.):
            warnings.warn(
                "Poorly determined shot noise coefficient "
                "for index pair {} and {}.".format(mu, nu),
                category=SphericalCoefficientWarning
            )

        return M_mu_nu * shot_noise_integral

    @staticmethod
    def _gen_operable_indices(disc, subtype=None, above_from=None):

        if subtype == 'angular':
            operable_indices = [
                (ell, m)
                for ell in disc.degrees
                for m in range(- ell, ell + 1)
            ]
        elif subtype == 'radial':
            operable_indices = [
                (ell, n)
                for ell, nmax in zip(disc.degrees, disc.depths)
                for n in range(1, nmax + 1)
            ]
        else:
            operable_indices = [
                (ell, m, n)
                for ell, nmax in zip(disc.degrees, disc.depths)
                for m in range(- ell, ell + 1)
                for n in range(1, nmax + 1)
            ]

        if above_from is not None:
            operable_indices = [
                index for index in operable_indices if index >= above_from
            ]

        return operable_indices



"""
***************************************************************************
Algorithms (:mod:`~harmonia.algorithms`)
***************************************************************************

Provide algorithms for Fourier basis evaluation and integration, spectrum
discretisation and structured array manipulation for cosmological data.

"""
from .arrays import CartesianArray, SphericalArray
from .discretisation import DiscreteSpectrum



"""
Synthesis (:mod:`~harmonia.surveyor.synthesis`)
===========================================================================

Systhesise FKP-style paired random catalogues for given survey
specifications.

This can be used for determining survey window functions and estimating
correlation induced by geometric filtering.

.. autosummary::

    SyntheticCatalogue
    CovarianceEstimator
    generate_compression_matrix

|

"""
import warnings

import numpy as np
import psutil
from scipy.interpolate import InterpolatedUnivariateSpline as Spline
from mcfit import P2xi
from nbodykit.lab import FKPCatalog

from harmonia.algorithms.arrays import CartesianArray
from harmonia.mapper.catalogue_maker import (
    RandomCatalogue,
    SphericalFKPCatalogue,
)
from harmonia.mapper.map_transform import CartesianMap, SphericalMap
from harmonia.reader.likelihoods import spherical_covariance


class SyntheticCatalogue:
    r"""Synthetic catalogue for probing survey geometry.

    Parameters
    ----------
    mean_density : float
        Mean number density (in cubic :math:`h`/Mpc).
    boxsize : float, array_like
        Catalogue boxsize (in Mpc/:math:`h`) as a scalar or a triple of
        scalars.
    contrast : float or None, optional
        Contrast of the number density of the secondary random catalogue
        to that of the primary.  If `None` (default), the secondary
        catalogue is not produced.  This cannot be `None` if
        `sphericalise` is provided.
    expansion : float, optional
        Expansion factor of the catalogue box for grid assignment
        (default is 1.).  Ignored if `sphericalise` is set.
    sphericalise : float or None, optional
        If not `None` (default), this is passed to the `radius` parameter
        of :class:`~.mapper.catalogue_maker.SphericalFKPCatalogue` for
        instantiating the synthetic catalogue.
    mask, selection, weight : callable or None, optional
        Survey mask, selection or weight function (default is `None`).
        Must be given as a function of Cartesian coordinate only assuming
        the origin is at the centre of the catalogues.
    apply_selection_as_veto : bool, optional
        If `True` (default is `False`), `selection` returning boolean
        values is applied to the 'Selection' column of the catalogues
        instead of the 'NZ' column (see `nbodykit` documentation for
        this peculiarity.)

    Attributes
    ----------
    spherical_coefficients : :class:`~.arrays.SphericalArray` or None
        Density contrast coefficients.  Only available after
        :meth:`compute_spherical_coefficients` is called.
    power_multipoles : :class:`~.algorithms.arrays.CartesianArray` or None
        Power spectrum multipoles.  Only available after
        :meth:`compute_power` is called.
    window_multipoles : :class:`~.algorithms.arrays.CartesianArray` or None
        Power spectrum multipoles of the window function at sampled
        separations.  Only available after :meth:`compute_window` is
        called.
    mask_multipoles : :class:`~.algorithms.arrays.CartesianArray` or None
        Correlation function multipoles of the window function at sampled
        separations.  Only available after :meth:`compute_mask` is
        called.
    attrs : dict
        Attributes inherited from input parameters.

    """

    def __init__(self, number_density, boxsize, contrast=None, expansion=1.,
                 sphericalise=None, mask=None, selection=None, weight=None,
                 apply_selection_as_veto=False):

        boxsize = [boxsize,] * 3 if np.isscalar(boxsize) \
            else boxsize

        self.attrs = {
            'number_density': number_density,
            'boxsize': boxsize,
            'contrast': contrast,
            'expansion': expansion,
            'sphericalise': sphericalise,
        }

        self._synthetic_catalogue = self._synthesise(
            number_density, boxsize, contrast, expansion,
            mask, selection, weight, apply_selection_as_veto,
            radius=sphericalise
        )

        self.spherical_coefficients = None
        self.power_multipoles = None
        self.window_multipoles = None
        self.mask_multipoles = None

        self._intrpl_k = None
        self._intrpl_power = None
        self._intrpl_s = None
        self._intrpl_correlation = None

    def __str__(self):

        str_info = ", ".join(
            [f"{name}={val}" for name, val in self.attrs.items()]
        )

        return f"{self.__class__.__name__}({str_info})"

    def compute_spherical_coefficients(self, disc):
        """Compute the spherical Fourier density contrast coefficients of
        the synthetic catalogue.

        Parameters
        ----------
        disc : :class:`~.algorithms.discretisation.DiscreteSpectrum`
            Discrete spectrum.

        Returns
        -------
        :class:`~harmonia.algorithms.arrays.SphericalArray`
            Density contrast coefficients.

        See Also
        --------
        :class:`~harmonia.mapper.map_transform.SphericalMap`

        """
        spherical_map = SphericalMap(self._synthetic_catalogue, disc)

        return spherical_map.density_contrast

    def compute_power(self, orders, kmin=1.e-4, kmax=None, dk=None,
                      num_mesh=256, resampler='tsc', interlaced=True):
        """Compute the power spectrum multipoles of the synthetic
        catalogue.

        Parameters
        ----------
        orders : sequence of int
            Orders of the power spectrum multipoles.
        kmin : float, optional
            Minimum wavenumber of the map (in :math:`h`/Mpc)
            (default is 1.e-4).
        kmax : float or None, optional
            Maximum wavenumber of the map (in :math:`h`/Mpc)
            (default is `None`).  If `None`, `kmax` is the Nyquist
            wavenumber determined from `num_mesh`.
        dk : float or None, optional
            Wavenumber bin width (in :math:`h`/Mpc) (default is `None`).
        num_mesh : int, optional
            Mesh number per dimension for interpolating the discrete
            catalogues on a grid.
        resampler : {'cic', 'tsc', 'pcs'}, optional
            Grid assignment scheme (default is ``'tsc'``) for catalogue
            interpolation.
        interlaced : bool, optional
            If `True` (default), use interlacing for aliasing mitigation.

        Returns
        -------
        :class:`~harmonia.algorithms.arrays.CartesianArray`
            Power spectrum multipoles binned in wavenumbers.

        See Also
        --------
        :class:`~harmonia.mapper.map_transform.CartesianMap`

        """
        if self.power_multipoles is not None:
            warnings.warn(
                "Power spectrum multipoles have already been computed; "
                "they are now being recomputed."
            )

        # Check if there is sufficient memory; if not,
        # reduce the mesh number.
        while psutil.virtual_memory().available < 10 * 768**3:
            num_mesh /= 2
            warnings.warn(
                "Mesh number reduced due to limited memory", RuntimeWarning
            )

        # Set the Nyquist wavenumber as ther upper limit.
        if kmax is None:
            kmax = np.pi * num_mesh \
                / min(self._synthetic_catalogue.attrs['BoxSize'])

        cartesian_map = CartesianMap(
            self._synthetic_catalogue,
            orders, kmin=kmin, kmax=kmax, dk=dk,
            num_mesh=num_mesh, resampler=resampler, interlaced=interlaced
        )

        with warnings.catch_warnings():
            # Filter out warnings when the secondary catalogue
            # in the synthetic FKP-style catalogue pair is None,
            # as it would cause division by zero.
            warnings.filterwarnings(
                'ignore',
                category=RuntimeWarning,
                message=".*invalid value encountered.*"
            )
            warnings.filterwarnings(
                'ignore',
                category=RuntimeWarning,
                message="divide by zero encountered in double_scalars"
            )
            self.power_multipoles = cartesian_map.power_multipoles

        return self.power_multipoles

    def compute_window(self, orders, kmin=1.e-6, **kwargs):
        """Compute the Fourier-space window function multipoles from the
        synthetic catalogue.

        Parameters
        ----------
        orders : sequence of int
            Orders of the power spectrum multipoles.
        kmin : float, optional
            Minimum wavenumber (in :math:`h`/Mpc) (default is 1.e-6).
        **kwargs
            Any other parameters to be passed to :meth:`~.compute_power`.

        Returns
        -------
        :class:`~harmonia.algorithms.arrays.CartesianArray`
            Window function multipoles binned in logarithmic wavenumbers.

        See Also
        --------
        :meth:`~.compute_power`

        """
        LOG10_K_MAX = 1.
        NUM_EXTENSION = 1000
        NUM_INTERPOLATION = pow(2, 10)

        if np.isclose(self.attrs['expansion'], 1.):
            warnings.warn(
                "No padding for catalogue box when computing "
                "the window function.  You may want to resynthesise "
                "the catalogue with `expansion` set above 1."
            )

        # Need to compute power spectrum multipoles first.
        power_multipoles = self.compute_power(orders, kmin=kmin, **kwargs)

        raw_k = power_multipoles.attrs['wavenumbers']
        raw_power = {
            ell: power_multipoles.array['power'][
                power_multipoles.array['order'] == ell
            ]
            for ell in orders
        }

        # Linear padding and then logarithmic padding in wavenumbers.
        padding = np.mean(np.abs(np.diff(raw_k)))
        lin_extension = np.max(raw_k) + padding * np.arange(1, NUM_EXTENSION)
        log_extension = np.logspace(
            np.log10(np.max(raw_k) + padding * NUM_EXTENSION),
            LOG10_K_MAX,
            num=NUM_EXTENSION
        )

        ext_k = np.r_[raw_k, lin_extension, log_extension]
        ext_power = {
            ell: np.r_[raw_power[ell], np.zeros(2 * NUM_EXTENSION - 1)]
            for ell in orders
        }

        # Resample at log-spaced wavenumbers.
        self._intrpl_k = np.logspace(
            *np.log10(ext_k[[0, -1]]), num=NUM_INTERPOLATION
        )
        self._intrpl_power = {
            ell: Spline(ext_k, ext_power[ell], k=1)(self._intrpl_k)
            for ell in orders
        }

        # Store as an attribute.
        orders = np.sort(orders).tolist()

        self.window_multipoles = CartesianArray(orders, self._intrpl_k)
        self.window_multipoles[:] = np.concatenate(
            [self._intrpl_power[ell] for ell in orders]
        )

        # Normalise all values to the monopole at vanishing wavenumbers.
        self.window_multipoles[:] /= self.window_multipoles[0]

        return self.window_multipoles

    def compute_mask(self, orders, *args, **kwargs):
        """Compute the configuration-space mask function multipoles from
        the synthetic catalogue by Hankel transform of the window function
        multipoles.

        Parameters
        ----------
        orders : sequence of int
            Orders of the mask function multipoles.
        *args, **kwargs
            Any other position and keyword parameters to be passed to
            :meth:`~.compute_window`.  Only required when
            :attr:`window_multipoles` is not available.

        Returns
        -------
        :class:`numpy.ndarray`
            Mask function multipoles binned in logarithmic separations.
            This is a NumPy structured array similar to
            :attr:`array` of :class:`~.algorithms.arrays.CartesianArray`.

        See Also
        --------
        :class:`~.algorithms.arrays.CartesianArray`

        """
        NUM_INTERPOL = pow(2, 14)

        # Need to compute window function multiples first.
        if self.window_multipoles is None \
                or max(orders) not in self._intrpl_power:
            self.window_multipoles = self.compute_window(
                orders, *args, **kwargs
            )

        # Hankel transform window function multipoles.
        with warnings.catch_warnings():  # suppress warnings from `mcfit`
            warnings.filterwarnings(
                'ignore',
                message=(
                    "The default value of extrap has been changed to False, "
                    "set it to True if you cannot reproduce previous results"
                )
            )
            raw_correlation = {
                ell: P2xi(self._intrpl_k, l=ell, lowring=True)(
                    self._intrpl_power[ell], extrap=False
                )
                for ell in orders
            }

        raw_s = [raw_correlation[ell][0] for ell in orders]
        raw_correlation.update({ell: raw_correlation[ell] for ell in orders})

        # Resample at log-spaced separations.
        self._intrpl_s = np.logspace(
            np.log10(np.min(raw_s)), np.log10(np.max(raw_s)), num=NUM_INTERPOL
        )
        self._intrpl_correlation = {
            ell: Spline(*raw_correlation[ell], k=1)(self._intrpl_s)
            for ell in orders
        }

        # Create a structure array dedicated for storing mask multipoles.
        self.mask_multipoles = np.empty(
            np.size(orders) * np.size(self._intrpl_s),
            dtype=np.dtype({
                'names': ['order', 'separation', 'correlation'],
                'formats': ['i4', 'f8', 'f8'],
            })
        )

        self.mask_multipoles['order'] = np.repeat(orders, len(self._intrpl_s))
        self.mask_multipoles['separation'] = np.tile(
            self._intrpl_s, len(orders)
        )

        # Normalise all values to the monopole at vanishing separations.
        self.mask_multipoles['correlation'] = np.concatenate(
            [self._intrpl_correlation[ell] for ell in orders]
        ) / raw_correlation[0][-1][0]

        return self.mask_multipoles

    @staticmethod
    def _synthesise(number_density, boxsize, contrast, expansion,
                    mask, selection, weight, apply_selection_as_veto,
                    radius=None):

        primary_catalogue = RandomCatalogue(number_density, boxsize)

        if contrast is not None:
            secondary_catalogue = RandomCatalogue(
                contrast * number_density, boxsize
            )
        else:
            secondary_catalogue = None

        # No sphericalisation; reimplement FKP build similar to
        # :class:`~harmonia.mapper.catalogue_maker.SphericalFKPCatalogue`.
        if radius is None:
            for catalogue in [primary_catalogue, secondary_catalogue]:
                if catalogue is None:
                    break

                catalogue['Location'] = \
                    catalogue['Position'] - np.divide(boxsize, 2)
                catalogue['NZ'] = number_density * catalogue['Selection']

                if callable(mask):
                    catalogue['Selection'] *= mask(catalogue['Location'])
                if callable(selection):
                    if apply_selection_as_veto:
                        catalogue['Selection'] *= \
                            selection(catalogue['Location'])
                    else:
                        catalogue['NZ'] *= selection(catalogue['Location'])
                if callable(weight):
                    catalogue['Weight'] *= weight(catalogue['Location'])

            return FKPCatalog(
                primary_catalogue, secondary_catalogue,
                BoxSize=np.multiply(expansion, boxsize)
            )

        # Reuse :class:`~.mapper.catalogue_maker.SphericalFKPCatalogue`.
        spherical_catalogue = SphericalFKPCatalogue(
            radius,
            data_catalogue=primary_catalogue,
            random_catalogue=secondary_catalogue,
            mask=mask, selection=selection, weight=weight,
            apply_selection_as_veto=apply_selection_as_veto
        )

        return FKPCatalog(
            spherical_catalogue.data_catalogue,
            spherical_catalogue.random_catalogue,
            BoxSize=np.multiply(expansion, boxsize)
        )


class CovarianceEstimator:
    """Covariance matrix estimator for power spectrum multipoles at
    fiducial values.

    Parameters
    ----------
    realisations : sequence of :class:`~.algorithms.arrays.CartesianArray`
        Independent realisations of power spectrum multipoles from which
        the covariance matrix is estimated.  For each realisation, orders
        and wavenumbers of the multipoles must be sorted.
    reference : :class:`~.arrays.CartesianArray` *or None, optional*
        Underlying power spectrum multipoles binned in wavenumber that
        are realised for covariance estimation.  Orders and wavenumbers of
        the multipoles must be sorted.  If `None` (default), this is
        determined by the average of `realisations`.

    Attributes
    ----------
    realisations : sequence of :class:`~.algorithms.arrays.CartesianArray`
        Independent realisations of power spectrum multipoles from which
        the covariance matrix is estimated.
    reference : :class:`~.arrays.CartesianArray` or None
        Underlying power spectrum multipoles binned in wavenumber that
        are realised for covariance estimation.
    wavenumbers : float :class:`ndarray`
        Wavenumbers at which the covariance matrix is estimated.
        This is set by `reference` if available; otherwise it is set by
        the first of `realisations`.

    See Also
    --------
    :class:`~harmonia.algorithms.arrays.CartesianArray`

    """

    def __init__(self, realisations, reference=None):

        self.realisations = realisations
        self.reference = reference

        if self.reference is not None:
            self.wavenumbers = np.unique(self.reference.array['wavenumber'])
        else:
            self.wavenumbers = np.unique(
                self.realisations[0].array['wavenumber']
            )

    def __getstate__(self):

        state = self.__dict__

        if self.reference is not None:
            state.update({'reference': self.reference.__getstate__()})

        return state

    def __setstate__(self, state):

        for attr, value in state.items():
            if attr == 'reference' and value is not None:
                reference = object.__new__(CartesianArray)
                reference.__setstate__(value)
                setattr(self, 'reference', reference)
            else:
                setattr(self, attr, value)

    def save(self, output_file):
        """Save the estimator with its attributes as a .npz file.

        Parameters
        ----------
        output_file : *str or* :class:`pathlib.Path`
            Output file path.

        """
        np.savez(output_file, **self.__getstate__())

    @classmethod
    def load(cls, input_file):
        """Load the estimator with its attributes from a .npz file.

        Parameters
        ----------
        input_file : *str or* :class:`pathlib.Path`
            Input file path.

        """
        state_data = np.load(input_file, allow_pickle=True)

        state = {}
        for attr in state_data.files:
            try:
                state.update({attr: state_data[attr].item()})
            except (AttributeError, ValueError):
                state.update({attr: state_data[attr]})

        self = object.__new__(cls)

        self.__setstate__(state)

        return self

    def get_fiducial_vector(self, pivot):
        """Return the fiducial data vector for which the covariance
        matrix is estimated.

        Parameters
        ----------
        pivot : {'order', 'wavenumber'}
            Pivot order used for data vectorisation.

        Returns
        -------
        float :class:`numpy.ndarray`
            Fiducial vector.

        """
        data_matrix = [
            realisation.vectorise(pivot) for realisation in self.realisations
        ]

        data_mean = np.mean(data_matrix, axis=0)

        if self.reference is None:
            return data_mean

        fiducial_vector = self.reference.vectorise(pivot)

        if not np.allclose(data_mean, fiducial_vector, rtol=0.1):
            warnings.warn(
                "Realisation average deviates from `truth` "
                "by more than 10% for some data entries."
            )

        return fiducial_vector

    def get_fiducial_covariance(self, pivot):
        """Return the fiducial covariance matrix estimated from data
        realisations.

        Parameters
        ----------
        pivot : {'order', 'wavenumber'}
            Pivot order used for data vectorisation.

        Returns
        -------
        float :class:`numpy.ndarray`
            Estimate fiducialcovariance.

        """
        data_matrix = [
            realisation.vectorise(pivot) for realisation in self.realisations
        ]

        return np.cov(data_matrix, rowvar=False, ddof=1)


def generate_compression_matrix(fiducial_model_kwargs,
                                extremal_model_kwargs=None,
                                sensitivity_threshold=0.01, discard=None):
    r"""Generate a compression matrix for spherical modes.

    Notes
    -----
    Compression is achieved by discarding non-positive eigenvalue modes
    that are at least :math:`10^{-8}` times smaller than the largest and
    in addition any of the following means:

        * `discard` is passed to discard a number of low-eigenvalue modes;
        * `extremal_model_kwargs` is passed and eigenvalues of the
          resulting model covariance are compared with those from
          `fiducial_covariance`.  Modes corresponding to low, insensitive
          (i.e. relative difference less than `sensitivity_threshold`)
          are discarded.
        * A combination of the above if the appropriate parameters
          are passed.

    Parameters
    ----------
    fiducial_model_kwargs : dict
        Fiducial model parameters to be passed to
        :func:`~.reader.likelihoods.spherical_covariance`.
    extremal_model_kwargs : dict or None, optional
        Extremal model parameters to be passed to
        :func:`~.reader.likelihoods.spherical_covariance`.
    sensitivity_threshold: float, optional
        Sensitivity threshold for modes deemed discardable
        (default is 0.01).
    discard : int or None, optional
        Number of low-eigenvalue modes to discard from all modes
        (default is `None`).

    Returns
    -------
    compression_matrix : :class:`numpy.ndarray`
        Compression matrix.

    """
    fiducial_covariance = spherical_covariance(**fiducial_model_kwargs)

    evals_fiducial, evecs = np.linalg.eigh(fiducial_covariance)

    selectors = []

    # Compression by positive magnitude.
    selectors.append(evals_fiducial > 1.e-8 * np.max(evals_fiducial))

    # Compression by discard.
    if discard is not None:
        selectors.append(np.arange(len(evals_fiducial)) >= discard)

    # Compression by comparison for sensitivity.
    extremal_covariance = spherical_covariance(**extremal_model_kwargs)

    evals_extremal = np.linalg.eigvalsh(extremal_covariance)

    selectors.append(
        ~np.isclose(evals_extremal, evals_fiducial, rtol=sensitivity_threshold)
    )

    # pylint: disable=no-member
    # Compress and reverse order.
    evecs = evecs[:, np.logical_and.reduce(selectors)][:, ::-1]

    compression_matrix = np.conj(evecs).T

    return compression_matrix



import numpy as np
import pytest
from mpi4py.MPI import COMM_WORLD

import harmonia.utils as u


@pytest.mark.parametrize(
    "data_array,mapping,comm,values",
    [
        ([1, 2, 3, 4], np.square, COMM_WORLD, [1, 4, 9, 16]),
    ]
)
def test_mpi_compute(data_array, mapping, comm, values):
    assert u.mpi_compute(data_array, mapping, comm=comm) \
        == pytest.approx(values), "MPI computation failed."


@pytest.mark.parametrize(
    "const,args",
    [
        (1j, (5, 'hello', None, False)),
    ]
)
def test_const_function(const, args):
    assert u.const_function(const)(*args) == pytest.approx(const), \
        "Constant function fails to return the specified constant."


@pytest.mark.parametrize(
    "vector,result",
    [
        (
            [[1, 0, 0], [1, 1, 0]],
            [[1, 0, 0], [1 / np.sqrt(2), 1 / np.sqrt(2), 0]]
        ),
    ]
)
def test_normalise_vector(vector, result):
    assert np.allclose(u.normalise_vector(vector), result), \
        "Vector normalisation failed."


@pytest.mark.slow
@pytest.mark.parametrize(
    "func,a,b,maxnum,values",
    [
        (np.sin, -0.01, 3.15, 1, np.array([0])),
        (np.cos, -0.01, 3.15, 10, np.array([np.pi/2])),
    ]
)
def test_binary_search(func, a, b, maxnum, values):
    assert u.binary_search(func, a, b, maxnum=maxnum) \
        == pytest.approx(values), \
        "Binary search fails to find all roots in the specified interval."


@pytest.mark.parametrize(
    "covar,value",
    [
        ([[2, 1], [1, 2]], [[1, 0.5], [0.5, 1]])
    ]
)
def test_covar_to_corr(covar, value):
    assert np.allclose(u.covar_to_corr(covar), value), \
        "Covariance conversion to correlation matrix failed."


@pytest.mark.parametrize(
    "matrix,value",
    [
        (np.array([[1, 2], [3, 4]]), 0.693147)
    ]
)
def test_mat_logdet(matrix, value):
    assert u.mat_logdet(matrix) == pytest.approx(value), \
        "Incorrect logarithmic determinant of matrix."


@pytest.mark.parametrize(
    "matrix,value",
    [
        (np.array([[1, 2], [3, 4]]), False)
    ]
)
def test_is_positive_definite(matrix, value):
    assert (u.is_positive_definite(matrix) is value), \
        "Positive definiteness check failed."



import numpy as np
import pytest
from nbodykit.cosmology import Planck15

from harmonia.surveyor.coordinates import (
    cartesian_to_spherical,
    sky_to_spherical,
    spherical_to_cartesian,
    spherical_to_sky,
    to_box_coords,
)


@pytest.mark.parametrize(
    "vec,value",
    [
        (
            [[1, 3, -5], [0.2, -0.88, -10]],
            [
                [5.916079783, 2.577650012, 1.249045772],
                [10.04063743, 3.051592333, 4.935866174],
            ]
        ),
    ]
)
def test_cartesian_to_spherical(vec, value):
    assert np.allclose(cartesian_to_spherical(vec), value), \
        "Incorrect coordinate conversion from Cartesian to spherical."


@pytest.mark.parametrize(
    "vec,value",
    [
        (
            [
                [5.916079783, 2.577650012, 1.249045772],
                [10.04063743, 3.051592333, 4.935866174],
            ],
            [[1, 3, -5], [0.2, -0.88, -10]]
        ),
    ]
)
def test_spherical_to_cartesian(vec, value):
    assert np.allclose(spherical_to_cartesian(vec), value), \
        "Incorrect coordinate conversion from Cartesian to Cartesian."


@pytest.mark.parametrize(
    "sky_coords,z_to_r,values",
    [
        ([0., 90.,], None, [np.pi/2, np.pi/2]),
        ([0., 0., 180.], Planck15.comoving_distance, [0., np.pi/2, np.pi]),
    ]
)
def test_sky_to_spherical(sky_coords, z_to_r, values):
    assert np.allclose(sky_to_spherical(sky_coords, z_to_r), values), \
        "Incorrect coordinate transform from sky to spherical."


@pytest.mark.parametrize(
    "spherical_coords,z_from_r,values",
    [
        ([1., np.pi/2, np.pi/2], None, [0., 90.,]),
    ]
)
def test_spherical_to_sky(spherical_coords, z_from_r, values):
    assert np.allclose(spherical_to_sky(spherical_coords, z_from_r), values), \
        "Incorrect coordinate transform from spherical to sky."


@pytest.mark.parametrize(
    "native_coord_system,box_centre,conversion_kwargs,coords,values",
    [
        ('null', 1., None, [1., 1., 1.], 0),
        ('sky', None, dict(z_from_r=lambda r: r), [1., 1., 1.],
         np.sqrt(3) \
            + (90 - np.rad2deg(np.arcsin(np.sqrt(2/3)))) + np.rad2deg(np.pi/4)
         ),
    ]
)
def test_to_box_coords(native_coord_system, box_centre, conversion_kwargs,
                       coords, values):

    @to_box_coords(
        native_coord_system=native_coord_system,
        box_centre=box_centre,
        conversion_kwargs=conversion_kwargs
    )
    def test_func(coords):
        return np.sum(coords, axis=-1)

    assert np.allclose(test_func(coords), values), \
        "Incorrect coordinate transform to box coordinates."
